def get_urls(main_url):
    return {
        'contest_list': main_url.split('?', 1)[0][:-6]+'register',
        'contest_info': main_url + '&action=2',
        'summary': main_url + '&action=137',
        'submissions': main_url + '&action=140&all_runs=1',
        'protocol': main_url + '&action=37&run_id={run_id}',
        'source': main_url + '&action=91&run_id={run_id}',
        'submission': main_url + '&action=139&prob_id={prob_id}',
        'submit': main_url.split('?')[0],
        'standings': main_url + '&action=94',
        'start_virtual': main_url + '&action=144',
        'stop_virtual': main_url + '&action=145',
        'download_file': main_url + '&prob_id={prob_id}&action=194&file={filename}',
        'clars': main_url + '&action=142&all_clars=1',
        'read_clar': main_url + '&action=39&clar_id={clar_id}',
        'sid': main_url.split("SID=")[-1].split("&")[0],
    }



class AggressiveCacheBackend:
    @staticmethod
    def detect(url):
        return url.startswith('cached:')
    @staticmethod
    def login_type(url):
        import ejcli.http
        return ejcli.http.login_type(url[7:])
    def __init__(self, url, login, password, **kwargs):
        import ejcli.http
        self.parent = ejcli.http.login(url[7:], login, password, **kwargs)[0]
        self.task_list_cache = None
        self.problem_info_cache = {}
        self.compiler_list_cache = {}
        self.submission_protocol_cache = {}
        self.submission_status_cache = {}
        self.submission_source_cache = {}
        self.compile_error_cache = {}
        self.clars_cache = {}
    def tasks(self):
        q = self.task_list_cache
        if q != None: return q
        self.task_list_cache = self.parent.tasks()
        return self.task_list_cache
    def submission_protocol(self, id):
        try: return self.submission_protocol_cache[id]
        except KeyError: pass
        q = self.parent.submission_protocol(id)
        import ejcli.commands.astatus
        if q: self.submission_protocol_cache[id] = q
        return q
    def compile_error(self, id):
        try: return self.compile_error_cache[id]
        except KeyError: pass
        ans = self.parent.compile_error(id)
        if ans: self.submission_protocol_cache[id] = ans
        return ans
    def submission_source(self, id):
        try: return self.submission_source_cache[id]
        except KeyError: pass
        self.submission_source_cache[id] = self.parent.submission_source(id)
        return self.submission_source_cache[id]
    def do_action(self, id, *args):
        if id == 'flush_cache':
            self.task_list_cache = None
            self.problem_info_cache.clear()
            self.compiler_list_cache.clear()
            self.submission_results_cache.clear()
            self.submission_status_cache.clear()
            self.submission_source_cache.clear()
            self.clars_cache.clear()
        else:
            self.parent.do_action(id, *args)
    def compiler_list(self, prob_id):
        try: return self.compiler_list_cache[prob_id]
        except KeyError: pass
        self.compiler_list_cache[prob_id] = self.parent.compiler_list(prob_id)
        return self.compiler_list_cache[prob_id]
    def submission_stats(self, subm_id):
        return self._submission_results(subm_id)[4]
    def problem_info(self, id):
        try: return self.problem_info_cache[id]
        except KeyError: pass
        self.problem_info_cache[id] = self.parent.problem_info(id)
        return self.problem_info_cache[id]
    def read_clar(self, id):
        try: return self.clars_cache[id]
        except KeyError: pass
        self.clars_cache[id] = self.parent.read_clar(id)
        return self.clars_cache[id]
    def contest_list(self):
        if isinstance(self, str):
            import ejcli.http
            return ejcli.http.contest_list(self[7:])
        else:
            return self.parent.contest_list()
    def __getattr__(self, attr):
        if attr == 'parent':
            raise AttributeError(attr)
        return getattr(self.parent, attr)



import collections

task_t = collections.namedtuple('task_t', ('id', 'short_name', 'long_name'))
submission_t = collections.namedtuple('submission_t', ('id', 'task', 'status', 'score', 'oktests'))
test_t = collections.namedtuple('test_t', ('status', 'stats'))
compiler_t = collections.namedtuple('compiler_t', ('id', 'short_name', 'long_name'))
clar_t = collections.namedtuple('clar_t', ('id', 'subject'))




import base64, json, collections
import ejcli._http.types as bjtypes
from .oauth import do_oauth
from ..base import Backend
from ..ejudge import get, post
from ...error import EJError
from ..html2md import html2md

def b64encode(s):
    return base64.b64encode(s).replace(b'=', b'').replace(b'/', b'_').replace(b'+', b'-')

def b64decode(s):
    s = s.replace(b'-', b'+').replace(b'_', b'/')
    while len(s) % 4: s += b'='
    return base64.b64decode(s)

class GCJ(Backend):
    @staticmethod
    def _detect(url):
        url = url.split('/')
        return len(url) > 4 and url[:3] == ['https:', '', 'codingcompetitions.withgoogle.com'] and url[4] == 'round'
    @staticmethod
    def detect(url):
        return url.startswith('https://codingcompetitions.withgoogle.com/')
    @staticmethod
    def login_type(url):
        if url.count('/') == 3: return ['contest_list']
        else: return ['goauth:email profile openid https://www.googleapis.com/auth/codejam']
    def __init__(self, url, login, password, token=None):
        Backend.__init__(self)
        if not self._detect(url):
            raise EJError("Invalid GCJ URL supplied")
        self.round = url.split('#', 1)[0].split('/')[5]
        if not set(self.round).issubset('0123456789abcdef'):
            raise EJError("Invalid GCJ URL supplied")
        self.token = token if token != None else do_oauth(url, login, password)
#       self.email = login #not used
        self._get_cache = {}
        code, headers, data = self._json_req('https://codejam.googleapis.com/dashboard/%s/poll'%self.round, {})
        if code != 200:
            raise EJError("Invalid contest ID")
    def _cache_get(self, url):
        with self.cache_lock:
            if url in self._get_cache:
                return self._get_cache[url]
        ans = get(url, {'Authorization': 'Bearer '+self.token})
        with self.cache_lock:
            if self.caching: self._get_cache[url] = ans
        return ans
    def _json_req(self, url, data, *, eout=True, ein=True, do_post=False):
        if eout: data = b64encode(json.dumps(data).encode('utf-8')).decode('ascii')
        if do_post: code, headers, data = post(url, {'p': data}, {'Authorization': 'Bearer '+self.token})
        else: code, headers, data = self._cache_get(url+'?p='+data)
        if ein: data = json.loads(b64decode(data).decode('utf-8', 'replace'))
        return code, headers, data
    def _get_which(self, which, error=None, data={}):
        code, headers, data = self._json_req('https://codejam.googleapis.com/%s/%s/poll'%(which, self.round), data)
        if code != 200 and error:
            raise EJError(error)
        return data
    def _debug_req(self, which, data={}):
        import pprint, pydoc
        pydoc.pager(pprint.pformat(self._json_req('https://codejam.googleapis.com/%s/%s/poll'%(which, self.round), data)))
    def _get_dashboard(self, q):
        ans = self._get_which('dashboard', q)
        ans['challenge']['tasks'].sort(key=lambda x: x['title'])
        ans['challenge']['tasks'].sort(key=lambda x: sum(i['value'] for i in x['tests']))
        return ans
    def tasks(self):
        data = self._get_dashboard("Failed to fetch task list.")
        return [bjtypes.task_t(i, hex(i)[2:], None) for i in (int(i['id'], 16) for i in data['challenge']['tasks'])]
    def submissions(self):
        data = self._get_which('attempts', "Failed to fetch submission list.")
        ans = []
        for i in data['attempts']:
            subm_id = int(i['id'], 16)
            task_id = hex(int(i['task_id'], 16))[2:]
            status = self._submission_status(i)
            score = self._submission_stats(data, i)[0]['score']
            try: oktests = sum(j['verdict__str'] == 'CORRECT' for j in data['judgement']['results'])
            except KeyError: oktests = None
            ans.append(bjtypes.submission_t(subm_id, task_id, status, score, oktests))
        return ans
    @staticmethod
    def _convert_verdict(s):
        s = s.replace('_', ' ')
        s = s[:1].upper()+s[1:].lower()
        if s == 'Correct':
            return 'OK'
        return s
    @staticmethod
    def _convert_stats(i):
        ans = {}
        if 'running_time_nanos' in i:
            ans['time_usage'] = i['running_time_nanos'] / 1000000000
        if 'running_memory_mbs' in i:
            ans['memory_usage'] = i['running_memory_mbs'] * 1048576
        return ans
    def _submission_descr(self, subm_id):
        subm_id = int(subm_id)
        data = self._get_which('attempts', "Failed to fetch submission list.")
        for i in data['attempts']:
            if int(i['id'], 16) == subm_id:
                return i
        return None
    def submission_protocol(self, subm_id):
        subm = self._submission_descr(subm_id)
        if subm == None: return []
        try: subm = subm['judgement']['results']
        except KeyError: return []
        return [bjtypes.test_t(self._convert_verdict(i['verdict__str']), self._convert_stats(i)) for i in subm]
    def submit_solution(self, task, lang, code):
        if isinstance(code, bytes): code = code.decode('utf-8', 'replace')
        self._json_req('https://codejam.googleapis.com/dashboard/%s/submit'%self.round, {'code': code, 'language_id': lang, 'task_id': '%016x'%task}, do_post=True)
        with self.cache_lock: self.stop_caching()
    def status(self):
        user_data = {}
        data = self._get_which('scoreboard', "Failed to fetch scoreboard.", {'num_consecutive_users': 0})
        for i in data['user_scores'][0]['task_info']:
            user_data[i['task_id']] = i['tests_definitely_solved'] + i['tests_possibly_solved']
        ans = collections.OrderedDict()
        for i in data['challenge']['tasks']:
            i_ = hex(int(i['id'], 16))[2:]
            st = None
            if i['id'] in user_data:
                st = 'OK' if user_data[i['id']] == sum(1 if j['value'] else 0 for j in i['tests']) else 'Partial solution'
            ans[i_] = st
        return ans
    def scores(self):
        user_data = {}
        data = self._get_which('scoreboard', "Failed to fetch scoreboard.", {'num_consecutive_users': 0})
        for i in data['user_scores'][0]['task_info']:
            user_data[i['task_id']] = i['score']
        ans = collections.OrderedDict()
        for i in data['challenge']['tasks']:
            i_ = hex(int(i['id'], 16))[2:]
            sc = None
            if i['id'] in user_data:
                sc = user_data[i['id']]
            ans[i_] = sc
        return ans
    def compile_error(self, subm_id):
        data = self._submission_descr(subm_id)
        if data == None: return None
        try: return data['judgement']['compilation_output']
        except KeyError: return None
    def _submission_status(self, subm):
        try: return self._convert_verdict(subm['judgement']['results'][-1]['verdict__str'])
        except KeyError: return 'Pending judgement'
    def submission_source(self, subm_id):
        data = self._submission_descr(subm_id)
        if data == None: return None
        return data['src_content'].encode('utf-8')
    def do_action(self, *args):
        raise EJError("Actions are not supported on GCJ.")
    def compiler_list(self, task):
        return [bjtypes.compiler_t(i['id'], i['id__str'].lower(), i['name']) for i in self._get_dashboard("Failed to fetch compiler list.")['languages']]
    def _submission_stats(self, data, subm):
        for task in data['challenge']['tasks']:
            if int(task['id'], 16) == int(subm['task_id'], 16):
                break
        else: task = {}
        ans = {'score': 0}
        ans['tests'] = {'total': 0, 'success': 0, 'fail': 0}
        for i, j in zip(subm.get('judgement', {}).get('results', ()), task.get('tests', 0)):
            if i['verdict__str'] == 'CORRECT': ans['score'] += j['value']
            if j['value'] != 0:
                ans['tests']['success' if i['verdict__str'] == 'CORRECT' else 'fail'] += 1
                ans['tests']['total'] += 1
        return (ans, None)
    def submission_stats(self, subm_id):
        data = self._get_which('attempts')
        for subm in data['attempts']:
            if int(subm['id'], 16) == subm_id: break
        else: return ({}, '')
        return self._submission_stats(data, subm)
    def contest_info(self):
        db = self._get_dashboard("Failed to fetch contest info.")
        data1 = {'contest_start': db['challenge']['start_ms'] / 1000, 'contest_end': db['challenge']['end_ms'] / 1000}
        data1['contest_duration'] = data1['contest_end'] - data1['contest_start']
        datas = {'Contest start time': time.ctime(data1['contest_start']), 'Duration': time.ctime(data1['contest_duration'])}
        return ('', datas, data1)
    def problem_info(self, task):
        for i in self._get_dashboard("Failed to fetch problem info.")['challenge']['tasks']:
            if int(i['id'], 16) == task:
                return ({}, html2md(i['statement']))
    def stop_caching(self):
        self._get_cache.clear()
    def contest_list(self):
        if not isinstance(self, str): return []
        competition = self.rsplit('/', 1)[-1]
        code, headers, data = get('https://codejam.googleapis.com/poll?p=e30')
        if code != 200: raise EJError("Failed to fetch contest list.")
        data = json.loads(b64decode(data).decode('utf-8', 'replace'))
        contests = [(j['start_ms'], j['title'], j['id']) for i in data['adventures'] if i['competition__str'].replace('_', '').lower() == competition for j in i['challenges']]
        contests.sort()
        return [(self+'/round/'+k, j, {}) for i, j, k in contests]



import time, threading, sys, io
from ejcli.hook_stdio import RedirectSTDIO

class Watch(threading.Thread):
    def __init__(self, bj, interval, cmd):
        self.bj = bj
        self.interval = interval
        self.cmd = cmd
        self.running = True
        self.prev_out = self.exec()
        threading.Thread(target=self.run).start()
    def exec(self):
        stdin = io.BytesIO()
        stdout = io.BytesIO()
        with RedirectSTDIO(stdin, stdout, stdout):
            self.bj.onecmd(self.cmd)
        return stdout.getvalue().decode('utf-8', 'replace')
    def run(self):
        while True:
            time.sleep(self.interval)
            if not self.running: break
            cur = self.exec()
            if cur != self.prev_out:
                print('Note: `'+self.cmd+'` changed!\r')
                if cur.startswith(self.prev_out):
                    print('(append)\r')
                    print(cur[len(self.prev_out):].replace('\n', '\r\n'))
                else:
                    print('<<<<<<< Was\r')
                    print(self.prev_out.replace('\n', '\r\n'))
                    print('=======\r')
                    print(cur.replace('\n', '\r\n'))
                    print('<<<<<<< Now\r')
                self.prev_out = cur

def do_watch(self, args):
    """
    usage: watch [options] [command]

        watch [-l <interval>] <command>
            Run <command> each <interval> (default=10) seconds and notify if the output changes.

        watch -d <watchno>
            Delete (cancel) watchpoint <watchno>. You will no longer receive notifications for that watchpoint.
    """
    if not hasattr(self, 'watches'):
        self.watches = {}
        self.watchno = 1
    if args.startswith('-d '):
        sp = args.split(' ')
        if len(sp) != 2 or not sp[1].isnumeric():
            return self.do_help('watch')
        idx = int(sp[1])
        if idx not in self.watches:
            print('No such watch:', idx)
        self.watches[idx].running = False
        del self.watches[idx]
    else:
        l = 10
        if args.startswith('-l '):
            sp = args.split(' ', 2)
            if len(sp) < 3 or not sp[1].isnumeric():
                return self.do_help('watch')
            l = int(sp[1])
            args = sp[2]
        args = args.strip()
        if args == '' or args.startswith('-'):
            return self.do_help('watch')
        self.watches[self.watchno] = Watch(self, l, args)
        print('Watch %d: %s'%(self.watchno, args))
        self.watchno += 1



import ejcli.http, getpass, ejcli.error

def do_passwd(self, cmd):
    """
    usage: passwd

    Change contest login password.
    """
    if cmd.strip():
        return self.do_help('passwd')
    oldpass = getpass.getpass('Old password: ')
    newpass = getpass.getpass('New password: ')
    newpass2 = getpass.getpass('Retype new password: ')
    if newpass != newpass2:
        raise ejcli.error.EJError('Passwords do not match!')
    ejcli.http.change_password(self.url, self.cookie, oldpass, newpass)
    print('Password changed successfully, now relogin.')
    self.url = self.cookie = None



import sys

def read_file(path, options=set()):
    with open(path) as file: return file.read()

def format(orig, options=set()):
    data = orig.split('\n')
    if '+pep8' in options:
        i = 0
        while i < len(data):
            if (data[i].strip().startswith("def ") or data[i].strip().startswith("class ")) and data[i - 1:i] != ['']:
                data.insert(i, '')
            if (data[i].startswith("def ") or data[i].startswith("class ")) and data[i - 2:i] != ['', '']:
                data.insert(i, '')
            i += 1
        while data[:1] == ['']: del data[0]
    data = '\n'.join(data)
    return data



from ejcli._http.ejudge import Ejudge, do_http, get, post
from ejcli._http.ejudge.ej371 import get_urls
from ejcli.error import EJError
from ejcli._http.base import Backend
import json, base64, collections, math
import ejcli._http.types as bjtypes

def mbjson(x):
    try: return json.loads(x)
    except json.JSONDecodeError: return None

def mb_problem_status(x):
    try: x = x.decode('utf-8')
    except UnicodeDecodeError: return None
    x = x.split('\n')
    add_later = {}
    y = []
    for i in x:
        #if i.startswith('      "input_file": '):
        #    add_later['input_file'] = i[20:-1]
        #elif i.startswith('      "output_file": '):
        #    add_later['output_file'] = i[21:-1]
        #else:
        y.append(i)
    y = '\n'.join(y)
    try: y = json.loads(y)
    except json.JSONDecodeError: return None
    y.update(add_later)
    return y

STATUS_NAMES = {i: j for i, j in enumerate(('OK', 'Compilation error', 'Run-time error', 'Time-limit exceeded', 'Presentation error', 'Wrong answer', 'Check failed', 'Partial solution', 'Accepted for testing', 'Ignored', 'Disqualified', 'Pending check', 'Memory limit exceeded', 'Security violation', 'Style violation', 'Wall time-limit exceeded', 'Pending review', 'Rejected', 'Skipped', 'Synchronization error'))}
STATUS_NAMES[23] = 'Summoned for defense'
STATUS_NAMES[95] = 'Full rejudge'
STATUS_NAMES[96] = 'Running...'
STATUS_NAMES[97] = 'Compiled...'
STATUS_NAMES[98] = 'Compiling...'
STATUS_NAMES[99] = 'Available for testing'

class EJFuse(Ejudge):
    @staticmethod
    def detect(url):
        return url.startswith('ejfuse:')
    def __init__(self, url, login, password):
        Backend.__init__(self)
        assert url.startswith('ejfuse:http://') or url.startswith('ejfuse:https://')
        url = url[7:].replace('/new-register?', '/new-client?').replace('/new-client?', '/client?')
        contest_id = url.split("contest_id=")[1].split("&")[0]
        self.contest_id = int(contest_id)
        base_url = url.split("/client?")[0]
        code, headers, data = post(base_url+'/register', {'action': 'login-json', 'login': login, 'password': password, 'json': 1})
        data = mbjson(data)
        if code != 200 or not data or not data['ok']:
            raise EJError("Login failed.")
        self.url = base_url+'/register'
        self.cookies = (data['result']['SID'], data['result']['EJSID'])
        code, headers, data = post(self.url, {'SID': self.cookies[0], 'EJSID': self.cookies[1], 'action': 'enter-contest-json', 'contest_id': self.contest_id, 'json': 1})
        data = mbjson(data)
        if code != 200 or not data or not data['ok']:
            raise EJError("Login failed.")
        self.url = base_url+'/client'
        self.cookies = (data['result']['SID'], data['result']['EJSID'])
        # will fall back to normal ejudge if an unimplemented feature is encountered
        self.urls = get_urls(base_url+'/new-client?SID='+self.cookies[0])
        self.cookie = 'EJSID='+self.cookies[1]
        self._get_cache = {}
    def tasks(self):
        code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=contest-status-json&json=1'%self.cookies, False)
        data = mbjson(data)
        if code != 200 or not data or not data['ok']:
            raise EJError("Failed to fetch task list.")
        return [bjtypes.task_t(i['id'], i['short_name'], i['long_name']) for i in data['result']['problems']]
    def _submission_list(self):
        code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=list-runs-json&prob_id=0&json=1'%self.cookies, False)
        data = mbjson(data)
        if code != 200 or not data or not data['ok']:
            raise EJError("Failed to fetch submission list.")
        return data['result']['runs']
    def submissions(self):
        data = self._submission_list()
        ti, tl, tll = list(zip(*self.tasks()))
        ti = {j:i for i, j in enumerate(ti)}
        return [bjtypes.submission_t(i['run_id'], tl[ti[i['prob_id']]], STATUS_NAMES[i['status']], i.get('score', None), None) for i in data]
    def _submission_descr(self, run_id):
        code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=run-status-json&run_id=%%d&json=1'%self.cookies%int(run_id), False)
        data = mbjson(data)
        if code != 200 or not data or not data['ok']:
            raise EJError("Failed to fetch testing protocol.")
        return data['result']
    def submission_protocol(self, run_id):
        data = self._submission_descr(run_id)
        try: data = data['testing_report']['tests']
        except KeyError: return [], []
        return [bjtypes.test_t(STATUS_NAMES[i['status']], {'time_usage': i['time_ms'] / 1000}) for i in data]
#   def submit(self, task, lang, text):
#       if isinstance(text, str): text = text.encode('utf-8')
#       try: task = self.task_ids()[task]#task += 1
#       except IndexError: return
#       data = []
#       data.append(b'"SID"\r\n\r\n'+self.cookies[0].encode('ascii'))
#       data.append(b'"EJSID"\r\n\r\n'+self.cookies[1].encode('ascii'))
#       data.append(b'"prob_id"\r\n\r\n'+str(task).encode('ascii'))
#       data.append(b'"lang_id"\r\n\r\n'+str(lang).encode('ascii'))
#       data.append(b'"file"; filename="brute.txt"\r\nContent-Type'
#                   b': text/plain\r\n\r\n'+text)
#       data.append(b'"JSON"\r\n\r\n1')
#       import random
#       while True:
#           x = b'----------'+str(random.randrange(1, 1000000000)).encode('ascii')
#           for i in data:
#               if x in i: break
#           else: break
#       x = '-----------------------------850577185583170701784494929'
#       data = b'\r\n'.join(b'--'+x+b'\r\nContent-Disposition: form-data; name='+i for i in data)+b'\r\n--'+x+b'--\r\n'
#       return post(self.url, data, {'Content-Type': 'multipart/form-data; boundary='+x.decode('ascii')})
    def status(self):
        tli = self.tasks()
        sl = self._submission_list()
        ans = collections.OrderedDict()
        for task in tli:
            code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=problem-status-json&problem=%%d&json=1'%self.cookies%task.id, False)
            data = mb_problem_status(data)
            if code != 200 or not data or not data['ok']:
                raise EJError("Failed to fetch task list.")
            try: best_run = data['result']['problem_status']['best_run']
            except KeyError: ans[task.short_name] = None
            else:
                for j in sl:
                    if j['run_id'] == best_run:
                        st = STATUS_NAMES[j['status']]
                        if st not in ('OK', 'Compiling...', 'Running...', 'Compilation error', 'Pending review', 'Pending check', 'Rejected', 'Style violation'):
                            st = 'Partial solution'
                        ans[task.short_name] = st
                        break
                else: ans[task.short_name] = None
        return ans
    def scores(self):
        tli = self.tasks()
        ans = collections.OrderedDict()
        for task in tli:
            code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=problem-status-json&problem=%%d&json=1'%self.cookies%task.id, False)
            data = mb_problem_status(data)
            if code != 200 or not data or not data['ok']:
                raise EJError("Failed to fetch task list.")
            try: ans[task.short_name] = data['result']['problem_status']['best_score']
            except KeyError: ans[task.short_name] = None
        return ans
    def compile_error(self, subm_id, *, binary=False, kind=None):
        def _decode_bytes(x):
            x = base64.b64decode(x['content']['data'])
            if binary: return x
            return x.decode('utf-8', 'replace')
        subm = self._submission_descr(subm_id)
        if 'testing_report' in subm and 'valuer_comment' in subm['testing_report'] and kind in (None, 2):
            return _decode_bytes(subm['testing_report']['valuer_comment'])
        elif 'compiler_output' in subm and kind in (None, 1):
            return _decode_bytes(subm['compiler_output'])
        else:
            return None
    def submission_status(self, subm_id):
        sl = self._submission_list()
        for i in sl:
            if i['run_id'] == subm_id:
                return STATUS_NAMES[i['status']]
        return None
    def submission_source(self, subm_id):
        code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=download-run&run_id=%%d&json=1'%self.cookies%subm_id, False)
        if code != 200 or headers.get('Content-Type', None) == 'text/json': return None
        return data
#   def do_action(self, subm_id):
#       raise EJError("Not supported on ejudge-fuse")
    def compiler_list(self, prob_id):
        code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=contest-status-json&json=1'%self.cookies, False)
        data = mbjson(data)
        if code != 200 or not data or not data['ok']:
            raise EJError("Failed to fetch compiler list.")
        return [bjtypes.compiler_t(i['id'], i['short_name'], i['long_name']) for i in data['result']['compilers']]
    def contest_info(self):
        code, headers, data = self._cache_get(self.url+'?SID=%s&EJSID=%s&action=contest-status-json&json=1'%self.cookies, False)
        data = mbjson(data)
        if code != 200 or not data or not data['ok']:
            raise EJError("Failed to fetch contest info.")
        datas = {}
        data1 = {}
        ctst = data['result']['contest']
        ctst.update(data)
        #datas['Server time:'] = data1['server_time'] = data['server_time']
        data1['server_time'] = data['server_time']
        #datas['Contest start time'] = data1['contest_start'] = ctst['start_time']
        data1['contest_start'] = ctst['start_time']
        #datas['Duration:'] = data1['contest_duration'] = math.inf if ctst['is_unlimited'] else ctst['duration']
        data1['contest_duration'] = math.inf if ctst['is_unlimited'] else ctst['duration']
        if 'end_time' in data:
            data1['contest_end'] = data['end_time']
        elif 'contest_start' in data1 and 'contest_duration' in data1:
            data1['contest_end'] = data1['contest_start'] + data1['contest_duration']
        if 'server_time' in data1 and 'contest_start' in data1:
            data1['contest_time'] = data1['server_time'] - data1['contest_start']
        return ('', datas, data1)
    def submission_stats(self, subm_id):
        subm = self._submission_descr(subm_id)
        ans = {}
        if 'score' in subm['run']: ans['score'] = subm['run']['score']
        tests = {}
        if 'passed_tests' in subm['run']: tests['success'] = subm['run']['passed_tests']
        if 'tests' in subm: tests['total'] = len(subm['tests'])
        if 'total' in tests and 'success' in tests:
            tests['fail'] = tests['total'] - tests['success']
        if tests: ans['tests'] = tests
        return (ans, None)
    #TODO
    def submission_score(self, subm_id):
        for i in self._submission_list():
            if i['run_id'] == subm_id:
                return i.get('score', None)
        return None
    def get_samples(self, subm_id):
        return self._get_samples(Ejudge.compile_error(self, subm_id))



from ejcli.commands.asubmit import do_asubmit

def do_aformat(self, cmd):
    """
    usage: aformat [-x <extension>] <file>

    Prints a style-fixed version of <file> to standard output.
    Uses a specific style fixer if -x is specified.
    """
    do_asubmit(self, cmd, afmt=True)



import subprocess, json, urllib.request
from ...error import EJError

PHJS = r'''
var page = require('webpage').create();
page.open('https://accounts.google.com/o/oauth2/auth?redirect_uri=storagerelay%3A%2F%2Fhttps%2Fcodingcompetitions.withgoogle.com%3Fid%3Dauth459599&response_type=permission%20id_token&scope=email%20profile%20openid%20https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fcodejam&openid.realm=&client_id=786918852754-bprfsksl61u8t4iua13hkq56jpja9g2u.apps.googleusercontent.com&ss_domain=https%3A%2F%2Fcodingcompetitions.withgoogle.com&prompt=select_account&fetch_basic_profile=true&gsiwebsdk=2', function()
{
    page.evaluate(function()
    {
        document.getElementById('Email').value = EMAIL;
        document.getElementById('next').click();
    });
    var tout = setInterval(function()
    {
        var q = page.evaluate(function()
        {
            var elem = document.getElementById('errormsg_0_Email');
            if(elem && elem.innerHTML !== '')
                return "invalidEmail";
            var elem = document.getElementById('Passwd');
            if(elem)
            {
                elem.value = PASSWORD;
                document.getElementById('signIn').click();
                return true;
            }
            return false;
        });
        if(q === "invalidEmail")
        {
            console.log('"invalidEmail"');
            phantom.exit();
        }
        if(q)
        {
            clearInterval(tout);
            tout = setInterval(function()
            {
                var q = page.evaluate(function()
                {try{
                    if(document.location.href.substr(0, 42) === "https://accounts.google.com/o/oauth2/auth?" && document.documentElement.outerHTML.indexOf('var done = function() {window.close();}; lso.relayAuthResponse( ') >= 0)
                    {
                        var sp = document.documentElement.outerHTML.split('var done = function() {window.close();}; lso.relayAuthResponse( ');
                        sp.shift();
                        sp = sp.join('var done = function() {window.close();}; lso.relayAuthResponse( ').split(');</script>');
                        sp.pop();
                        sp = sp.join(');</script>').split('{');
                        sp[0] = '';
                        sp = sp.join('{').split('}');
                        sp[sp.length-1] = '';
                        sp = JSON.parse(sp.join('}'));
                        history.pushState({}, '', 'https://accounts.google.com/o/oauth2/iframe');
                        var xhr = new XMLHttpRequest();
                        xhr.open('GET', 'https://accounts.google.com/o/oauth2/iframerpc?action=issueToken&response_type=token%20id_token&login_hint='+sp['login_hint']+'&client_id=786918852754-bprfsksl61u8t4iua13hkq56jpja9g2u.apps.googleusercontent.com&origin=https%3A%2F%2Fcodingcompetitions.withgoogle.com&scope=openid%20profile%20email%20https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fcodejam&ss_domain=https%3A%2F%2Fcodingcompetitions.withgoogle.com', false);
                        xhr.setRequestHeader('X-Requested-With', 'XmlHttpRequest');
                        xhr.send('');
                        return xhr.responseText.split("\n").join(" ");
                    }
                    var elem = document.getElementById('errormsg_0_Passwd');
                    if(elem && elem.innerHTML !== '')
                        return '"invalidPassword"';
                    var elem = document.getElementById('submit_approve_access');
                    if(elem)
                        return '"noAccess"';
                    return "null";
                }catch(e){return ""+e;}});
                if(q !== "null")
                {
                    console.log(q);
                    clearInterval(tout);
                    phantom.exit();
                }
            }, 100);
        }
    }, 100);
});
'''

def do_oauth(email, password):
    phjs = PHJS.replace('EMAIL', json.dumps(email)).replace('PASSWORD', json.dumps(password))
    p = subprocess.Popen(('phantomjs', '/dev/stdin'), stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    res = json.loads(p.communicate(phjs.encode('utf-8'))[0].decode('utf-8', 'replace').split('\n', 1)[0])
    p.wait()
    if res == 'invalidEmail':
        raise EJError("Login failed: invalid e-mail.")
    elif res == 'invalidPassword':
        raise EJError("Login failed: invalid password.")
    elif res == 'noAccess':
        raise EJError("Login failed: rights haven't been granted. Log in to GCJ using your web browser, then try again.")
    return res['access_token']



def get_urls(main_url):
    if '?SID=' in main_url:
        main_url, sid = main_url.split('?SID=')
        sid = 'S' + sid
    else:
        main_url, sid = main_url.rsplit('/', 1)
        sid = sid.split('?')[0]
    if main_url.endswith('/main-page'):
        main_url = main_url[:-10]
    if main_url.endswith('/register'):
        main_url = main_url[:-9] + '/user'
    return {
        'contest_list': main_url[:-5]+'/register',
        'contest_info': main_url + '/main-page/' + sid,
        'summary': main_url + '/view-problem-summary/' + sid,
        'submissions': main_url + '/view-submissions/' + sid + '?all_runs=1',
        'protocol': main_url + '/view-report/' + sid + '?run_id={run_id}',
        'source': main_url + '/download-run/' + sid + '?run_id={run_id}',
        'submission': main_url + '/view-problem-submit/' + sid + '?prob_id={prob_id}',
        'submit': main_url,
        'standings': main_url + '/standings/' + sid,
        'start_virtual': main_url + '?SID=' + sid[1:] + '&action=144',
        'stop_virtual': main_url + '?SID=' + sid[1:] + '&action=145',
        'download_file': main_url + '?SID=' + sid[1:] + '&prob_id={prob_id}&action=194&file={filename}',
        'clars': main_url + '/view-clars/' + sid + '?all_clars=1',
        'read_clar': main_url + '/view-clar/' + sid + '?clar_id={clar_id}',
        'sid': sid[1:]
    }



import json, base64, socket, urllib.parse
from ejcli.error import EJError
from ejcli._http.base import Backend
from ejcli._http.ejudge import do_http, get, post
import ejcli._http.types as bjtypes

def urlescape(s):
    return urllib.parse.urlencode({'x': s})[2:]

def json_req(url, data, headers={}, method=None):
    if data != None:
        data = json.dumps(data).encode('utf-8')
        if method == None: method = 'POST'
    else:
        data = b''
        if method == None: method = 'GET'
    code, headers, data = do_http(url, method, headers, data)
    try: data = json.loads(data.decode('utf-8'))
    except (json.JSONDecodeError, UnicodeDecodeError): data = None
    return (code, headers, data)

class JJS(Backend):
    @staticmethod
    def detect(url):
        sp = url.split('/')
        return len(sp) >= 2 and sp[0] in ('http+jjs:', 'https+jjs:') and not sp[1]
    @staticmethod
    def login_type(url):
        if '?' in url:
            url, params = url.split('?')
            params = {k: v for k, v in (i.split('=', 1) if '=' in i else (i, None) for i in params.split('&'))}
        else:
            params = {}
        if 'contest' not in params:
            return ['contest_list']
        elif params.get('auth', None) == 'token':
            return ['pass']
        elif params.get('auth', None) in ('gettoken', 'guest'):
            return []
        else:
            return ['login', 'pass']
    @staticmethod
    def _get_uuid(subm_id):
        uu = '%032x'%subm_id
        return uu[:8]+'-'+uu[8:12]+'-'+uu[12:16]+'-'+uu[16:20]+'-'+uu[20:32]
    def __init__(self, url, login, password):
        Backend.__init__(self)
        url, params = url.split('?')
        url = url.replace('+jjs', '', 1)
        if url.endswith('/'): url = url[:-1]
        params = {k: v for k, v in (i.split('=', 1) if '=' in i else (i, None) for i in params.split('&'))}
        contest_id = params['contest']
        if params.get('auth', None) == 'token':
            self.cookie = 'Bearer '+password
        elif params.get('auth', None) == 'gettoken':
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.connect('/tmp/jjs-auth-sock')
            token = b''
            while True:
                chk = sock.recv(1024)
                token += chk
                if not chk: break
            token = token.decode('ascii').strip()
            if not (token.startswith('===') and token.endswith('===')):
                raise EJError('Login failed: failed to get token from jjs-auth-sock')
            self.cookie = 'Bearer '+token[3:-3]
        elif params.get('auth', None) == 'guest':
            self.cookie = 'Bearer Guest'
        else:
            code, headers, data = json_req(url + '/auth/simple', {'login': login, 'password': password})
            if code != 200:
                try: msg = 'Login failed: %s'%data['detail']
                except Exception: msg = 'Login failed.'
                raise EJError(msg)
            self.cookie = 'Bearer '+data['data']
        self.url = url
        self.params = params
        self.contest = contest_id
        self.lsu_cache = {}
        self._get_cache = {}
        code, headers, data = json_req(url + '/contests/' + urlescape(self.contest), None, {'Authorization': self.cookie})
        if code != 200 or data == None:
            raise EJError('Login failed: unknown contest')
    def _cache_get(self, path):
        try: return self._get_cache[path]
        except KeyError: pass
        ans = json_req(self.url + path, None, {'Authorization': self.cookie})
        with self.cache_lock:
            if self.caching:
                self._get_cache[path] = ans
        return ans
    def stop_caching(self):
        self._get_cache.clear()
    def tasks(self):
        code, headers, data = self._cache_get('/contests/'+urlescape(self.contest)+'/problems')
        if code != 200: return []
        return [bjtypes.task_t(i, j['rel_name'], j['name']) for i, j in enumerate(data)]
    def submissions(self):
        code, headers, data = self._cache_get('/contests/'+urlescape(self.contest)+'/problems')
        if code != 200: mapping = {}
        else: mapping = {i['name']: i['rel_name'] for i in data}
        code, headers, data = self._cache_get('/runs')
        if code != 200: return []
        data.reverse()
        return [bjtypes.submission_t(
            int(i['id'].replace('-', ''), 16),
            mapping.get(i['problem_name'], i['problem_name']),
            self._submission_status(lsu, i),
            self._submission_status(lsu, i),
            lsu.get('test', None) if lsu != None else None
        ) for i, lsu in ((i, self._get_lsu(i['id'])) for i in data if i['contest_name'] == self.contest)]
    def submission_protocol(self, id):
        code, headers, data = self._cache_get('/runs/%s/protocol?compile_log=true&resource_usage=true'%self._get_uuid(int(id)))
        if code != 200: return []
        return [bjtypes.test_t(self._format_status(i['status']['code']), {'time_usage': i['time_usage']/1000000000, 'memory_usage': i['memory_usage']}) for i in data['tests']]
    def submit_solution(self, taskid, lang, text):
        tl = self.tasks()
        if taskid not in range(len(tl)): return
        cl = self.compiler_list(taskid)
        taskid = tl[taskid][1]
        if lang not in range(len(cl)): return
        lang = cl[lang][1]
        if isinstance(text, str): text = text.encode('utf-8')
        code, headers, data = json_req(self.url+'/runs', {
            'toolchain': lang,
            'code': ''.join(base64.b64encode(text).decode('ascii').split()),
            'problem': taskid,
            'contest': self.contest
        }, {'Authorization': self.cookie})
        if code != 200:
            try: msg = 'Submit failed: '+data['detail']
            except: return
            else: raise EJError(msg)
        with self.cache_lock: self.stop_caching()
    def status(self):
        ans = {}
        with self.may_cache():
            data = list(zip(self.task_list(), self.scoreboard()[0][1]))
        for i, j in data:
            if j == None: ans[i] = None
            elif j['attempts'] < 0: ans[i] = 'Partial solution'
            else: ans[i] = 'OK'
        return ans
    def scores(self):
        with self.may_cache():
            data = list(zip(self.task_list(), self.scoreboard()[0][1]))
        return {i: (j['score'] if j != None else None) for i, j in data}
    def compiler_list(self, task):
        code, headers, data = self._cache_get('/toolchains')
        if code == 200:
            return [bjtypes.compiler_t(i, x['id'], x['description']) for i, x in enumerate(data)]
        else:
            raise EJError("Failed to fetch compiler list.")
    def _submission_descr(self, id):
        id = self._get_uuid(int(id))
        code, headers, data = self._cache_get('/runs')
        if code != 200: return None
        for i in data:
            if i['id'] == id:
                return i
    def _get_lsu(self, id):
        return None
        id = self._get_uuid(int(id))
        code, headers, lsu = self._cache_get('/runs/%d/live'%id)
        if code != 200:
            return self.lsu_cache.get(id, None)
        if lsu['finish']:
            try: del self.lsu_cache[id]
            except KeyError: pass
            return None
        with self.cache_lock:
            if id not in self.lsu_cache: self.lsu_cache[id] = {'test': None, 'score': None}
            if lsu['current_test'] != None: self.lsu_cache[id]['test'] = lsu['current_test']
            if lsu['live_score'] != None: self.lsu_cache[id]['score'] = lsu['live_score']
            return self.lsu_cache[id]
    def _format_status(self, st):
        if st == 'ACCEPTED' or st == 'TEST_PASSED' or st == 'OK': return 'OK'
        st = st.replace('_', ' ')
        return st[:1].upper()+st[1:].lower()
    def compile_error(self, id, *, binary=False, kind=None):
        id = self._get_uuid(int(id))
        if kind in (None, 1): # compiler output
            code, headers, data = self._cache_get('/runs/%s/protocol?compile_log=true&resource_usage=true'%id)
            if code != 200: ans = None
            else: ans = base64.b64decode(data.get('compile_stdout', '').encode('ascii'))+base64.b64decode(data.get('compile_stderr', '').encode('ascii'))
        elif kind == 3: # binary
            code, headers, data = get(self.url+'/runs/%s/binary'%id, {'Authorization': self.cookie})
            if code != 200: ans = None
            else: ans = data
        if ans != None and not binary: ans = ans.decode('utf-8', 'replace')
        return ans
    def contest_info(self):
        code, headers, data = self._cache_get('/system/is-dev')
        if code != 200: return ('', {}, {})
        return ('', {}, {'jjs_devmode': data})
    def _submission_status(self, lsu, st):
        if lsu != None:
            status = 'Running'
            if lsu['test'] != None: status += ', test '+str(lsu['test'])
            return status
        if st == None: return None
        if not st['status']: return 'Running'
        return self._format_status(st['status']['code'])
    def submission_source(self, id):
        code, headers, data = self._cache_get('/runs/%s/source'%self._get_uuid(int(id)))
        if code != 200: return None
        return base64.b64decode(data.encode('ascii'))
    def submission_stats(self, id):
        lsu = self._get_lsu(id)
        if lsu != None:
            ans['score'] = lsu['score']
            if lsu['test'] != None: ans['tests'] = {'success': lsu['test']}
            return ans
        st = self._submission_descr(id)
        if st == None: return None
        ans = {}
        code, headers, prot = self._cache_get('/runs/%s/protocol?compile_log=true&resource_usage=true'%self._get_uuid(int(id)))
        if code != 200: prot = None
        if 'subtasks' in prot and prot['subtasks']:
            prot['subtasks'].sort(key=lambda i: i['subtask_id'])
            ans['group_scores'] = [i['score'] for i in prot['subtasks']]
        return (ans, '')
    def _submission_score(self, lsu, st):
        if lsu != None: return lsu['score']
        if st == None: return None
        return st['score']
    def get_samples(self, id, *, binary=False):
        def deb64(x):
            ans = base64.b64decode(x.encode('ascii'))
            if not binary: ans = ans.decode('utf-8', 'replace')
            return ans
        ans = {}
        code, headers, data = self._cache_get('/runs/%s/protocol?test_data=true&output=true&answer=true'%self._get_uuid(int(id)))
        if code != 200: return ans
        for i, j in enumerate(i for i in data['tests']):
            cur = ans[i + 1] = {}
            for k1, k2 in (('test_stdin', 'Input'), ('test_stdout', 'Output'), ('test_stderr', 'Stderr'), ('test_answer', 'Correct')):
                if k1 in j and j[k1] != None: cur[k2] = deb64(j[k1])
        return ans
    def scoreboard(self):
        code, headers, data = self._cache_get('/contests/'+urlescape(self.contest)+'/standings')
        if code != 200: return []
        ans = []
        i = 1
        while str(i) in data['parties']:
            cur = data['parties'][str(i)]
            i += 1
            ans.append(({'name': 'STUB', 'color': cur['stats']['color'], 'total_score': cur['stats']['score']}, []))
            j = 1
            while str(j) in cur['problems']:
                cur2 = cur['problems'][str(j)]
                j += 1
                if cur2['empty']: ans[-1][1].append(None)
                else:
                    ans[-1][1].append({'score': cur2['score'], 'attempts': cur2['attempts'] * (1 if cur2['ok'] else -1)})
        return ans
    def contest_list(self):
        if isinstance(self, str):
            if '?' in self:
                url, params = self.split('?')
                params = {k: v for k, v in (i.split('=', 1) if '=' in i else (i, None) for i in params.split('&'))}
            else:
                url = self
                params = {}
            if url.endswith('/'): url = url[:-1]
        else:
            url = self.url
            params = dict(self.params)
        url0 = url
        url = url.replace('+jjs', '', 1)
        code, headers, data = json_req(url+'/contests', None)
        ans = []
        if code != 200: return ans
        for i in data:
            params['contest'] = i['id']
            ans.append((url0+'/?'+urllib.parse.urlencode(params), i['title'], {}))
        return ans



import urllib.parse
from ..error import EJError

class FileBackend:
    @staticmethod
    def detect(url):
        return url.startswith('file:')
    def __new__(self, url, *args0, **kwds0):
        import imp
        path = url[5:].split('#', 1)[0]
        if '?' in path:
            path, args = path.split('?', 1)
            args = dict(urllib.parse.parse_qsl(args))
        else:
            args = {}
        path = urllib.parse.unquote(path)
        cls = args.get('class', 'TheBackend')
        try: file = open(path)
        except IOError:
            raise EJError("File not found.")
        mod = imp.load_module('file:'+path, file, path, ('.py', 'r', 1))
        if not hasattr(mod, cls):
            raise EJError("Backend class not found.")
        return getattr(mod, cls)(url, *args0, **kwds0)



import urllib.request, urllib.parse, html, json, ssl
from ...error import EJError

def parse_hidden_inputs(data):
    ans = {}
    for i in data.split('\n  <input ')[1:]:
        i = ' '+i.split('>', 1)[0]
        if ' type="hidden" ' in i:
            name = i.split(' name="', 1)[1].split('"', 1)[0]
            value = i.split(' value="', 1)[1].split('"', 1)[0]
            ans[html.unescape(name)] = html.unescape(value)
    if 'bgresponse' in ans:
        ans['bgresponse'] = 'js_enabled'
    return ans

def do_oauth(email, password):
    ctx = ssl.create_default_context()
    x = urllib.request.build_opener(urllib.request.HTTPCookieProcessor, urllib.request.HTTPRedirectHandler, urllib.request.HTTPSHandler(context=ctx))
    data1 = x.open('https://accounts.google.com/o/oauth2/auth?redirect_uri=storagerelay%3A%2F%2Fhttps%2Fcodingcompetitions.withgoogle.com%3Fid%3Dauth459599&response_type=permission%20id_token&scope=email%20profile%20openid%20https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fcodejam&openid.realm=&client_id=786918852754-bprfsksl61u8t4iua13hkq56jpja9g2u.apps.googleusercontent.com&ss_domain=https%3A%2F%2Fcodingcompetitions.withgoogle.com&prompt=select_account&fetch_basic_profile=true&gsiwebsdk=2')
    data1 = data1.read().decode('utf-8', 'replace')
    data1 = parse_hidden_inputs(data1)
    data1['Email'] = email
    data2 = x.open('https://accounts.google.com/signin/v1/lookup', urllib.parse.urlencode(data1).encode('ascii'))
    data2 = data2.read().decode('utf-8', 'replace')
    data2 = parse_hidden_inputs(data2)
    data2['Email'] = email
    data2['Passwd'] = password
    data3 = x.open('https://accounts.google.com/multilogin/challenge/sl/password', urllib.parse.urlencode(data2).encode('ascii'))
    data3 = data3.read().decode('utf-8', 'replace')
    if '<span role="alert" class="error-msg" id="errormsg_0_Passwd">' in data3:
        raise EJError("Login failed.")
    if '<img src="https://accounts.google.com/Captcha?' in data3:
        raise EJError("Login failed: CAPTCHA received. Log in to GCJ using your web browser, then try again.")
    if '<button id="submit_approve_access" ' in data3:
        raise EJError("Login failed: rights haven't been granted. Log in to GCJ using your web browser, then try again.")
    data4 = json.loads('{'+data3.split('lso.relayAuthResponse( ', 1)[1].split('{', 1)[1].rsplit(');</script>', 1)[0].rsplit('}', 1)[0]+'}')
#curl 'https://accounts.google.com/o/oauth2/iframerpc?action=issueToken&response_type=token%20id_token&login_hint=AJDLj6LJaXjplD11OecOt22tY_Tlr9x5xx7KVtECyntMGiQ1ovSeFwZ3tlhiSnD5UDm49mvfRC9tVxiuNRWi97jQrqnuOZ8Aa1p6U9HGJ9iBRG6VJvGwClU&client_id=786918852754-bprfsksl61u8t4iua13hkq56jpja9g2u.apps.googleusercontent.com&origin=https%3A%2F%2Fcodingcompetitions.withgoogle.com&scope=openid%20profile%20email%20https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fcodejam&ss_domain=https%3A%2F%2Fcodingcompetitions.withgoogle.com' -H 'X-Requested-With: XmlHttpRequest' -H 'Cookie: SID=lAfXdY0HfsLBLe-YXBoSWvZvoXXeqslGBFVfq9eD-E4RDVq100B3BYx_fV0HSldSeVaVlg.; LSID=ah|blogger|doritos|lso|mail|o.calendar.google.com|o.console.cloud.google.com|o.mail.google.com|o.myaccount.google.com|o.notifications.google.com|o.smartlock.google.com|o.takeout.google.com|s.RU|s.blogger|s.youtube|sitemaps|ss|urlshortener|wise|youtube:lAfXdezUGT-6s-fj7Y-y3EGoBlr9LaaBgVDIBB9x2g4CW21XHyrjvS4Ew-GZNYrF9F5SAQ.; HSID=AVYr8nELzElaznxtX; SSID=AvS_FHt-pndJPrbUt' -H 'Connection: keep-alive' -v; echo
    data5 = x.open(urllib.request.Request('https://accounts.google.com/o/oauth2/iframerpc?action=issueToken&response_type=token%20id_token&login_hint='+data4['login_hint']+'&client_id=786918852754-bprfsksl61u8t4iua13hkq56jpja9g2u.apps.googleusercontent.com&origin=https%3A%2F%2Fcodingcompetitions.withgoogle.com&scope=openid%20profile%20email%20https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fcodejam&ss_domain=https%3A%2F%2Fcodingcompetitions.withgoogle.com', headers={'X-Requested-With': 'XmlHttpRequest'}))
    data5 = json.loads(data5.read().decode('utf-8', 'replace'))
    return data5['access_token']



import sys
from ejcli.http import submissions
from ejcli.error import EJError

def still_running(s):
    return s.endswith('...') or ', ' in s or s in ('Compiling', 'Running', 'Judging', 'Check failed', 'Available for testing', 'Full rejudge', 'Pending check', 'Pending judgement', 'Queue judge', 'In queue', 'Running on pretest')

def do_astatus(self, cmd):
    """
    usage: astatus <subm_id>

    Fancy testing progress display
    """
    subm_id = cmd.strip() 
    if not subm_id.isnumeric():
        return self.do_help('astatus')
    chars = '\\|/-'
    idx = 0
    prev = ''
    subm_id = int(subm_id)
    while True:
        try:
            cur = next(i.status for i in submissions(self.url, self.cookie) if i.id == subm_id)
        except StopIteration:
            raise EJError('No such submission')
        cur = cur.strip()
        sys.stderr.write(' '*len(prev)+'\r')
        sys.stderr.flush()
        if still_running(cur):
            prev = '%%%ds' % len(prev) % (cur + ' ' + chars[idx])
            idx = (idx + 1) % 4
#           print(prev)
            sys.stderr.write(prev+'\r')
            sys.stderr.flush()
        else:
            print(cur)
            break



import shlex
from ejcli.http import tasks, clar_list, submit_clar, read_clar

def do_clars(self, cmd):
    """
    usage: clars <command> [args...]

    `command` can be one of:
        list
        List clars.

        send <task> <subject>
        Send clar from stdin.

        read <clar_id>
        Read clar.
    """
    cmd = shlex.split(cmd)
    if not cmd: cmd = ['list']
    if cmd[0] not in ('list', 'send', 'read'):
        return self.do_help('clars')
    elif cmd[0] == 'list':
        if len(cmd) != 1: return self.do_help('clars')
        print('Clar ID\tSubject')
        for k, v in clar_list(self.url, self.cookie):
            print('%d\t%s'%(k, v))
    elif cmd[0] == 'read':
        if len(cmd) != 2 or not cmd[1].strip().isnumeric(): return self.do_help('clars')
        print(read_clar(self.url, self.cookie, int(cmd[1])))
    elif cmd[0] == 'send':
        if len(cmd) != 3: return self.do_help('clars')
        try: task = next(i.id for i in tasks(self.url, self.cookie) if i.short_name == cmd[1])
        except StopIteration:
            raise EJError("No such task")
        data = ''
        while not data.endswith('\n\n'):
            try: data += input()+'\n'
            except EOFError: break
        submit_clar(self.url, self.cookie, task, cmd[2], data.strip())



import urllib.request, urllib.parse, html, json, time
from ejcli._http.base import Backend
from ejcli.error import EJError
from ejcli._http.openerwr import OpenerWrapper
import ejcli._http.html2md as html2md
import ejcli._http.types as bjtypes

class CodeForces(Backend):
    @staticmethod
    def detect(url):
        url = url.split('/')
        return url[0] in ('http:', 'https:') and not url[1] and ('.'+url[2]).endswith('.codeforces.com') and url[3] in ('contest', 'contests')
    @staticmethod
    def _get_csrf(data):
        return data.split('<meta name="X-Csrf-Token" content="', 1)[1].split('"', 1)[0]
    def __init__(self, url, login, password):
        Backend.__init__(self)
        self.locale = 'en'
        if '#locale=' in url:
            url, self.locale = url.rsplit('#locale=', 1)
        if url.startswith('http:'):
            url = 'https:' + url[5:]
        if url.find('/contest') == url.find('/contests'):
            url = '/contest'.join(url.split('/contests', 1))
        self.base_url = url
        self.handle = login
        self.host = url.split('/')[2]
        self.opener = OpenerWrapper(urllib.request.build_opener(urllib.request.HTTPCookieProcessor))
        if login != None or password != None:
            csrf = self._get_csrf(self.opener.open('https://%s/enter?back=%%2F'%self.host).read().decode('utf-8', 'replace'))
            ln = self.opener.open('https://%s/enter?back=%%2F'%self.host, urllib.parse.urlencode({
                'csrf_token': csrf,
                'action': 'enter',
                'ftaa': '',
                'bfaa': '',
                'handleOrEmail': login,
                'password': password
            }).encode('ascii'))
            if ln.geturl() != 'https://%s/'%self.host:
                raise EJError("Login failed.")
        self._gs_cache = None
        self._st_cache = None
        self._subms_cache = {}
    def _get_submit(self):
        data = self.opener.open(self.base_url+'/submit?locale=en')
        if data.geturl() not in (self.base_url+'/submit', self.base_url+'/submit?locale=en'):
            return [], [], None
        data = data.read().decode('utf-8', 'replace')
        csrf = self._get_csrf(data)
        data1 = data.split('name="submittedProblemIndex">', 1)[1].split('</select>', 1)[0].split('<option value="')
        tasks = [i.split('"', 1)[0] for i in data1[2:]]
        data2 = data.split('name="programTypeId">', 1)[1].split('</select>', 1)[0].split('<option value="')
        langs = [(int(i.split('"', 1)[0]), html.unescape(i.split('>', 1)[1].split('</option>', 1)[0].strip())) for i in data2[1:]]
        short_codes = {43: 'gcc', 42: 'g++11', 50: 'g++14', 54: 'g++', 2: 'msvc2010', 59: 'msvc2017', 9: 'mcs', 7: 'python', 31: 'python3', 40: 'pypy', 41: 'pypy3'}
        langs_ans = []
        for i, j in langs:
            langs_ans.append(bjtypes.compiler_t(i, short_codes.get(i, str(i)), j))
        return (tasks, langs_ans, csrf)
    def _get_submissions(self):
        with self.cache_lock:
            if self._gs_cache != None: return self._gs_cache
        data = self.opener.open(self.base_url+'/my?locale=en')
        if data.geturl() not in (self.base_url+'/my', self.base_url+'/my?locale=en'):
            raise EJError("Failed to fetch submission list")
        data = data.read().decode('utf-8', 'replace')
        csrf = self._get_csrf(data)
        data = data.replace('<tr class="last-row" data-submission-id="', '<tr data-submission-id="').split('<tr data-submission-id="')
        subms = []
        for i in data[1:]:
            subm_id = int(i.split('"', 1)[0])
            meta = {}
            data2 = i.split('>', 1)[1].split('</tr>', 1)[0].split('<td')
            for j in data2[1:]:
                try: cls = j.split('class="', 1)[1].split('"', 1)[0]
                except IndexError: cls = None
                data = j.split('>', 1)[1].split('</td>', 1)[0]
                meta[cls] = data
            subms.append((subm_id, meta))
        ans = (subms, csrf)
        with self.cache_lock:
            if self.caching: self._gs_cache = ans
        return ans
    def _get_submission(self, idx, csrf):
        if idx in self._subms_cache: return self._subms_cache[idx]
        req = self.opener.open('https://'+self.host+'/data/submitSource', urllib.parse.urlencode(
        {
            'submissionId': idx,
            'csrf_token': csrf
        }).encode('ascii'))
        ans = json.loads(req.read().decode('utf-8', 'replace'))
        if self.caching: self._subms_cache[idx] = ans
        return ans
    def tasks(self):
        q = self._get_submit()[0]
        if not q:
            data = self.opener.open(self.base_url).read().decode('utf-8')
            ans = []
            sp = data.split('<td class="id">\r\n                        <a href="'+self.base_url.rsplit('codeforces.com', 1)[1]+'/problem/')
            if sp[0].rfind('<') > sp[0].rfind('>'):
                return ans
            for i in sp[1:]:
                ans.append(i.split('"', 1)[0])
            return [bjtypes.task_t(i, j, None) for i, j in enumerate(ans)]
        return [bjtypes.task_t(i, j, None) for i, j in enumerate(q)]
    def submissions(self, all=False):
        if all:
            contest_id = self.base_url.split('/')
            while not contest_id[-1]: contest_id.pop()
            contest_id = int(contest_id[-1])
            data = json.loads(self.opener.open('https://'+self.host+'/api/contest.status?contestId=%d'%contest_id).read().decode('utf-8', 'replace'))
            if data['status'] != 'OK':
                raise EJError('Failed to load submissions.')
            return [bjtypes.submission_t(
                i['id'],
                i['problem']['index'],
                self._format_status(i['verdict']),
                None,
                i['passedTestCount']
            ) for i in data['result']]
        else:
            #TODO: use API here too
            data = self._get_submissions()[0]
            return [bjtypes.submission_t(
                i[0],
                i[1]['status-small'].split('<a href="', 1)[1].split('"', 1)[0].rsplit('/', 1)[1],
                self._format_total_status(i[1]['status-cell status-small status-verdict-cell'].split('>', 1)[-1]),
                None,
                self._get_oktests(i[1]['status-cell status-small status-verdict-cell'].split('>', 1)[-1])
            ) for i in data]
    @staticmethod
    def _format_status(st):
        if st == 'OK': return 'OK'
        st = st.replace('_', ' ')
        return st[:1].upper()+st[1:].lower()
    @staticmethod
    def _format_total_status(st):
        v = st.split('<')
        v = v[0]+''.join(i.split('>', 1)[1] for i in v[1:])
        v = v.split(' on test ', 1)[0]
        v = v.split(' on pretest ', 1)[0]
        v = v.split('&nbsp;(', 1)[0]
        v = v.strip()
        return 'OK' if v == 'Accepted' else v
    @staticmethod
    def _get_oktests(st):
        v = st.split('<')
        v = v[0]+''.join(i.split('>', 1)[1] for i in v[1:])
        v = v.split(' on test ', 1)[-1]
        v = v.split(' on pretest ', 1)[-1]
        q = v.split('&nbsp;(')
        if len(q) == 2:
            q[-1] = q[-1].strip()
            assert q[-1].endswith(')')
            try: return int(q[-1][:-1])
            except ValueError: pass
        v = v.strip()
        try: return int(v) - 1
        except ValueError: return None
    def submission_protocol(self, subm_id):
        data = self._get_submission(subm_id, self._get_submissions()[1])
        ntests = int(data.get('testCount', 0))
        ans = []
        for i in range(ntests):
            stats = {}
            try: stats['time_usage'] = float(data['timeConsumed#'+str(i+1)])/1000
            except KeyError: pass
            try: stats['memory_usage'] = int(data['memoryConsumed#'+str(i+1)])
            except KeyError: pass
            try: ans.append(bjtypes.test_t(self._format_status(data['verdict#'+str(i+1)]), stats))
            except KeyError: pass
        return ans
    def _submit(self, task, lang, text, csrf):
        if isinstance(text, str): text = text.encode('utf-8')
        data = []
        data.append(b'"ftaa"\r\n\r\n')
        data.append(b'"bfaa"\r\n\r\n')
        data.append(b'"action"\r\n\r\nsubmitSolutionFormSubmitted')
        data.append(b'"submittedProblemIndex"\r\n\r\n'+task.encode('utf-8'))
        data.append(b'"programTypeId"\r\n\r\n'+str(lang).encode('utf-8'))
        data.append(b'"source"\r\n\r\n'+text)
        import random
        while True:
            x = b'----------'+str(random.randrange(1, 1000000000)).encode('ascii')
            for i in data:
                if x in i: break
            else: break
        data = b'\r\n'.join(b'--'+x+b'\r\n'+b'Content-Disposition: form-data; name='+i for i in data)+b'\r\n--'+x+b'--\r\n'
        try:
            self.opener.open(urllib.request.Request(self.base_url+'/submit?csrf_token='+csrf,
                                data=data,
                                headers={'Content-Type': 'multipart/form-data; boundary='+x.decode('ascii')},
                                method='POST'))
        except urllib.request.URLError: pass
    def submit_solution(self, task, lang, text):
        tasks, langs, csrf = self._get_submit()
        self._submit(tasks[task], lang, text, csrf)
        with self.cache_lock: self.stop_caching()
    def status(self):
        subms = self._get_submissions()[0]
        ans = {j: None for i, j, k in self.tasks()}
        for i, j in subms:
            task = j['status-small'].split('<a href="', 1)[1].split('"', 1)[0].rsplit('/', 1)[1]
            status = self._format_total_status(j['status-cell status-small status-verdict-cell'].split('>', 1)[-1])
            if status in ('Accepted', 'Pretests passed'): status = 'OK'
            if ans.get(task, None) == None or status == 'OK': ans[task] = status
        return ans
    def scores(self):
        with self.cache_lock: data = self._st_cache
        if data == None:
            data = self.opener.open(self.base_url+'/standings?locale=en').read().decode('utf-8', 'replace')
            with self.cache_lock:
                if self.caching: self._st_cache = data
        tasks = (i.split('href="/contest/', 1)[1].split('"', 1)[0].rsplit('/', 1)[1] for i in data.split('<th ')[5:])
        for i in data.split('<tr participantId="')[1:]:
            i = i.split('</tr>', 1)[0]
            handle = i.split('<a href="/profile/', 1)[1].split('"', 1)[0]
            if handle != self.handle: continue
            ans = {}
            for j in i.replace('<td\r\n'+' '*16+'problemId="', '<td\r\n'+' '*16+'>').split('<td\r\n'+' '*16+'>')[1:]:
                j = j.split('<span class="cell-', 1)[1].split('>', 1)[1].split('</span>', 1)[0]
                try: j = int(j)
                except ValueError: j = -1
                ans[next(tasks)] = j if j >= 0 else None
            return ans
        return {}
    def scoreboard(self):
        contest_id = self.base_url.split('/')
        while not contest_id[-1]: contest_id.pop()
        contest_id = int(contest_id[-1])
        data = json.loads(self.opener.open('https://'+self.host+'/api/contest.standings?contestId=%d&from=1&count=1000000000'%contest_id).read().decode('utf-8', 'replace'))
        if data['status'] != 'OK':
            raise EJError('Failed to load scoreboard.')
        return [
            ({'name': ', '.join(j['handle'] for j in i['party']['members'])}, [
                ({'score': j['points'], 'attempts': j['bestSubmissionTimeSeconds']} if 'bestSubmissionTimeSeconds' in j else None)
            for j in i['problemResults']])
        for i in data['result']['rows']]
    def _compile_error(self, subm_id, csrf):
        return json.loads(self.opener.open('https://'+self.host+'/data/judgeProtocol',
            urllib.parse.urlencode({
                'submissionId': subm_id,
                'csrf_token': csrf
            }).encode('ascii')).read().decode('utf-8', 'replace'))
    def compile_error(self, subm_id):
        return self._compile_error(subm_id, self._get_submissions()[1])
    def submission_source(self, subm_id):
        subm = self._get_submission(subm_id, self._get_submissions()[1])
        if 'source' in subm: return subm['source'].encode('utf-8')
    def compiler_list(self, prob_id):
        return self._get_submit()[1]
    def submission_stats(self, subm_id):
        subm = self._get_submission(subm_id, self._get_submissions()[1])
        ans = {}
        if 'testCount' in subm and subm['testCount'] and int(subm['testCount']) != 0:
            ntests = int(subm['testCount'])
            ans['tests'] = {'total': ntests}
            success = 0
            for i in range(ntests):
                if subm.get('verdict#'+str(i+1), None) == 'OK':
                    success += 1
            ans['success'] = success
            ans['fail'] = ntests - success
        return (ans, None)
    def contest_info(self):
        ans = {}
        data = self.opener.open(self.base_url.replace('/contest/', '/contests/')+'?locale=en').read().decode('utf-8', 'replace')
        data = data.split('<span class="format-time"', 1)[1].split('>', 1)[1].strip()
        date = data.split('<', 1)[0] # dd.mm.yyyy hh:mm, Mmm/dd/yyyy hh:mm for English locale!!!
        duration = data.split('</span>\r\n                </a>\r\n    </td>\r\n    <td>', 1)[1].split('<', 1)[0].strip() # hh:mm
        q1 = {'Start': date, 'Length': duration}
        if ' class="countdown">' in data:
            countdown = data.split(' class="countdown">', 1)[1].split('<', 1)[0].strip() # hh:mm:ss
            q1['Time left'] = countdown
        else:
            countdown = None
        q2 = {}
        dur = 0
        assert duration.count(':') == 1
        for i in map(int, duration.split(':')):
            dur = 60 * dur + i
        dur *= 60
        if countdown is None:
            left = None
        else:
            left = 0
            assert countdown.count(':') == 2
            for i in map(int, countdown.split(':')):
                left = 60 * left + i
        date, tm = date.split()
        h, m = map(int, tm.split(':'))
        if '/' in date:
            mt, d, y = date.split('/')
            d = int(d)
            y = int(y)
            mt = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'].index(mt) + 1
        else:
            d, mt, y = map(int, date.split('.'))
        start = int(time.mktime((y, mt, d, h, m, 0, -1, -1, -1)))
        q2['contest_start'] = start
        q2['contest_duration'] = dur
        q2['contest_end'] = start + dur
        if left is not None:
            q2['contest_time'] = dur - left
            q2['server_time'] = start + (dur - left)
        return '', q1, q2
    def problem_info(self, prob_id):
        task = self.tasks()[prob_id][1]
        data = self.opener.open(self.base_url+'/problem/'+task+'?locale='+self.locale).read().decode('utf-8', 'replace')
        data = data.split('<div class="property-title">', 1)[1].split('</div><div>', 1)[1]
        data = data.split('<script>')[0]
        data = data.split('<script type="text/javascript">', 1)[0]
        return ({}, html2md.html2md(data, None, self.base_url+'/problems/'+task+'?locale='+self.locale))
    def download_file(self, *args):
        raise EJError("File download doesn't exist on CodeForces")
    def clar_list(self):
        raise EJError("Clarifications don't exits on CodeForces")
    def submit_clar(self, *args):
        raise EJError("Clarifications don't exits on CodeForces")
    def read_clar(self, id):
        raise EJError("Clarifications don't exits on CodeForces")
    def get_samples(self, subm_id):
        subm = self._get_submission(subm_id, self._get_submissions()[1])
        ans = []
        for i in range(int(subm.get('testCount', '0'))):
            cur = {}
            suf = '#%d'%(i+1)
            if 'input'+suf in subm:
                cur['Input'] = subm['input'+suf]
            if 'output'+suf in subm:
                cur['Output'] = subm['output'+suf]
            if 'answer'+suf in subm:
                cur['Correct'] = subm['answer'+suf]
            if 'checkerStdoutAndStderr'+suf in subm:
                cur['Checker output'] = subm['checkerStdoutAndStderr'+suf]
            ans.append(cur)
        return {i + 1: j for i, j in enumerate(ans)}
    def stop_caching(self):
        self._gs_cache = None
        self._st_cache = None
        self._subms_cache.clear()
    def contest_list(self):
        if isinstance(self, str):
            opener = urllib.request.build_opener()
        else:
            opener = self.opener
            self = 'https://codeforces.com/contests'
        if self.startswith('http:'): self = 'https:' + self[5:]
        data = opener.open(self).read().decode('utf-8').split('<tr\r\n    \r\n    data-contestId="')
        ans = []
        for i in data[1:]:
            cid, name = i.split('"', 1)
            cid = int(cid)
            name = html.unescape(name.split('<td>', 1)[1].split('<br/>', 1)[0].split('</td>', 1)[0].strip())
            ans.append((name, 'https://codeforces.com/contest/'+str(cid), {}))
        return ans
    def locales(self):
        return [('en', 'English'), ('ru', 'Russian')]
    def set_locale(self, which):
        self.locale = which



from ejcli.http import get_samples, has_feature
from ejcli.error import EJError
import os.path, sys

def do_samples(self, cmd):
    """
    usage: samples [--binary] <submission_id> (<dump_dir> || --input || --output || --correct || --stderr || --checker)

    Dump sample test cases from a submission. If a flag is specified, dump the corresponding files to standard output.
    """
    if len(cmd.split(' ')) < 2:
        return self.do_help("samples")
    cmd, ddir = cmd.split(' ', 1)
    kwargs = {}
    if cmd == '--binary':
        if not has_feature(self.url, self.cookie, 'get_samples', 'binary'):
            raise EJError("This system does not support --binary in samples.")
        kwargs['binary'] = True
        if len(ddir.split(' ')) < 2:
            return self.do_help("samples")
        cmd, ddir = ddir.split(' ', 1)
    if not cmd.isnumeric():
        raise EJError("Submission ID must be a number")
    tests = get_samples(self.url, self.cookie, int(cmd), **kwargs)
    flags = {'--input': 'Input', '--output': 'Output', '--correct': 'Correct', '--stderr': 'Stderr', '--checker': 'Checker output'}
    if ddir not in flags:
        os.makedirs(ddir)
    out_sep = False
    for k, v in tests.items():
        for what, data in v.items():
            if ddir in flags:
                if what == flags[ddir]:
                    if out_sep: print('#'*20)
                    else: out_sep = True
                    (sys.stdout.buffer if 'binary' in kwargs else sys.stdout).write(data)
                    (sys.stdout.buffer if 'binary' in kwargs else sys.stdout).flush()
                continue
            if what == 'Input':
                suff = ''
            elif what == 'Correct':
                suff = '.a'
            else:
                suff = '_'+what.replace(' ', '_')+'.txt'
            with open(os.path.join(ddir, '%02d' % k + suff), 'wb' if 'binary' in kwargs else 'w') as file:
                file.write(data)



import webbrowser, http.server, threading, json, shlex
from .._http.ejudge import get, post
from ..http import login, login_type
from ..error import EJError
from urllib.parse import urlencode, unquote

HTTP_RESPONSE = b'''\
<html>
<head>
<title>ejcli Google OAuth success</title>
</head>
<body>
<p>ejcli has successfully authenticated against Google OAuth servers. You may now close this window.</p>
</body>
</html>'''

class LoginCodeReceiver(http.server.BaseHTTPRequestHandler):
    def do_GET(self):
        code = unquote(self.path.split('?code=', 1)[1].split('&', 1)[0])
        self.server.login_code = code
        threading.Thread(target=self.server.shutdown).start()
        self.send_response(200)
        self.send_header('Content-Type', 'text/html; charset=utf-8')
        self.send_header('Content-Length', str(len(HTTP_RESPONSE)))
        self.end_headers()
        self.wfile.write(HTTP_RESPONSE)

def get_login_code(scope, client_id):
    srv = http.server.HTTPServer(('127.0.0.1', 0), LoginCodeReceiver)
    redirect_uri = 'http://%s:%d'%srv.socket.getsockname()
    webbrowser.open('https://accounts.google.com/o/oauth2/auth?'+urlencode({'redirect_uri': redirect_uri, 'scope': scope, 'response_type': 'code', 'client_id': client_id}))
    srv.serve_forever()
    return srv.login_code, redirect_uri

def get_auth_token(login_code, redirect_uri, client_id, client_secret):
    code, headers, data = post('https://oauth2.googleapis.com/token', {'code': login_code, 'client_id': client_id, 'client_secret': client_secret, 'redirect_uri': redirect_uri, 'grant_type': 'authorization_code'}, {'Content-Type': 'application/x-www-form-urlencoded'})
    assert code == 200, data
    return json.loads(data)['access_token']

def do_login(scope, client_id, client_secret):
    login_code, redirect_uri = get_login_code(scope, client_id)
    return get_auth_token(login_code, redirect_uri, client_id, client_secret)

def do_googlelogin(self, cmdline):
    """
    usage: googlelogin [--client-id <client_id>] [--client-secret <client_secret>] <url>

    Login to the testing system using Google OAuth (the testing system must support such logins).
    """
    cmdline = shlex.split(cmdline)
    client_id = '158805258134-fk2gef9qi727b7rncel01p582md374a0.apps.googleusercontent.com'
    client_secret = 'scWWMjrZmnQxA6BPz_4dYgxa'
    if cmdline and cmdline[0] == '--client-id':
        client_id = cmdline[1]
        del cmdline[:2]
    if cmdline and cmdline[0] == '--client-secret':
        client_secret = cmdline[1]
        del cmdline[:2]
    if len(cmdline) != 1:
        return self.do_help('googlelogin')
    url = cmdline[0]
    x = [i for i in login_type(url) if i.startswith('goauth:')]
    if len(x) != 1:
        raise EJError("Google OAuth is not supported by this testing system.")
    scope = x[0][7:]
    token = do_login(scope, client_id, client_secret)
    self.url, self.cookie = login(url, None, None, token=token)



import urllib.request, urllib.parse, html, json, random
from .base import Backend
from ..error import EJError
from .openerwr import OpenerWrapper

class YaContest(Backend):
    @staticmethod
    def detect(url):
        if url.endswith('/'): url = url[:-1]
        if url.endswith('/enter'): url = url[:-6]
        return url.startswith('https://contest.yandex.ru/contest/') and url[34:].isnumeric()
    @staticmethod
    def _get_bem(data, sp):
        return json.loads(html.unescape(data.split(sp, 1)[1].split('"', 1)[0]))
    @classmethod
    def _get_sk(self, data):
        try: return data.split('<input type="hidden" name="sk" value="', 1)[1].split('"', 1)[0]
        except IndexError: return self._get_bem(data, '<div class="aside i-bem" data-bem="')['aside']['sk']
    def __init__(self, url, login, passwd):
        if url.endswith('/'): url = url[:-1]
        if url.endswith('/enter'): url = url[:-6]
        if not self.detect(url):
            raise EJError("Not a contest.yandex.ru URL")
        self.opener = OpenerWrapper(urllib.request.build_opener(urllib.request.HTTPCookieProcessor))
        data = self.opener.open('https://passport.yandex.ru/auth?'+urllib.parse.urlencode({'origin': 'consent', 'retpath': 'https://passport.yandex.ru/profile'}), urllib.parse.urlencode({'login': login, 'passwd': passwd}).encode('ascii'))
        if data.geturl() != 'https://passport.yandex.ru/profile':
            print(data.geturl())
            raise EJError('Login failed.')
        self.url = url
        self.short_url = '/'+url.split('/', 3)[3]
    def task_list(self):
        data = self.opener.open(self.url+'/problems/').read().decode('utf-8', 'replace')
        return [html.unescape(i) for i in (i.split('/', 1)[0] for i in data.split('<a class="link" href="'+self.short_url+'/problems/')[1:]) if '"' not in i]
    def task_ids(self):
        return list(range(len(self.task_list())))
    def submit(self, task, lang, code):
        if isinstance(code, str): code = code.encode('utf-8')
        t = self.task_list()[task]
        data = self.opener.open(self.url+'/problems/'+t+'/').read().decode('utf-8', 'replace')
        prob_id = self._get_bem(data, '<div class="solution solution_type_compiler-list i-bem" data-bem="')['solution']['problemId']
        cmplrs = self._compiler_list(data)
        sk = self._get_sk(data)
        data = []
        data.append(b'')
        data.append(prob_id.encode('ascii')+b'@compilerId"\r\n\r\n'+cmplrs[lang][1].encode('ascii')+b'\r\n')
        data.append(prob_id.encode('ascii')+b'@solution"\r\n\r\ntext\r\n')
        data.append(prob_id.encode('ascii')+b'@text"\r\n\r\n'+code+b'\r\n')
        data.append(b'sk"\r\n\r\n'+sk.encode('ascii')+b'\r\n')
        data.append(b'retpath"\r\n\r\nhttps://ya.ru/\r\n')
        rand = b''
        while any(rand in i for i in data):
            rand = ('%020d'%random.randrange(10**20)).encode('ascii')
        data = (b'--'+rand+b'\r\nContent-Disposition: form-data; name="').join(data)+b'--'+rand+b'--\r\n'
        data2 = self.opener.open(urllib.request.Request(self.url+'/submit/', data, {'Content-Type': 'multipart/form-data; boundary='+rand.decode('ascii')}))
        if '?error=' in data2:
            raise EJError('Error: '+data2.split('?error=', 1)[1])
        with self.cache_lock: self.stop_caching()
    @staticmethod
    def _compiler_list(data):
        ans = []
        for i in data.split('<select class="select__control" name="')[1:]:
            if i.split('"', 1)[0].endswith('@compilerId'):
                for idx, j in enumerate(i.split('</select>', 1)[0].split('<option class="select__option" value="')[1:]):
                    short_name = html.unescape(j.split('"', 1)[0])
                    long_name = html.unescape(j.split('</option>', 1)[0].rsplit('>', 1)[1])
                    ans.append((idx, short_name, long_name))
        return ans
    def compiler_list(self, task):
        t = self.task_list()[task]
        data = self.opener.open(self.url+'/problems/'+t+'/').read().decode('utf-8', 'replace')
        return self._compiler_list(data)
    def do_action(self, action, *args):
        try: url = self.url + {'stop_virtual': '/finish/?return=false', 'restart_virtual': '/finish/?return=true'}[action]
        except KeyError: pass
        else:
            #WIP, doesn't work yet
            try: return self.opener.open(url, urllib.parse.urlencode({'sk': self._get_sk(self.opener.open(self.url).read().decode('utf-8', 'replace'))}).encode('ascii')).read() == b'OK'
            except urllib.request.URLError: return False
        action = {'register': 'register', 'start_virtual': 'startVirtual'}[action]
        try: data = self.opener.open(self.url+'/enter/', urllib.parse.urlencode({'sk': self._get_sk(self.opener.open(self.url).read().decode('utf-8', 'replace')), 'action': action, 'retpath': self.url}).encode('ascii'))
        except urllib.request.URLError: return False
        url = data.geturl()
        return url.startswith(self.url+'/') and '?error=' not in url



import subprocess
from ejcli.http import tasks, download_file
from ejcli.error import EJError

def do_getfile(self, cmd):
    """
    usage: getfile <prob_id> <file> <shell command>

    Retrieve a file attached to a task.
    """
    id, file, cmd = (cmd.strip()+'  ').split(' ', 2)
    if not id or not file:
        return self.do_help('getfile')
    try: task_id = next(i.id for i in tasks(self.url, self.cookie) if i.short_name == id)
    except StopIteration:
        raise EJError("No such task.")
    data = download_file(self.url, self.cookie, task_id, file)
    p = subprocess.Popen('cat '+cmd, stdin=subprocess.PIPE, shell=True)
    p.stdin.write(data)
    p.stdin.close()
    p.wait()



import sys, threading, io, subprocess, os

class TLDProxy:
     def __init__(self, default):
         object.__setattr__(self, '_default', default)
         object.__setattr__(self, '_tld', threading.local())
     @property
     def _the_tld(self):
         try: return self._tld.value
         except AttributeError: return self._default
     def __getattr__(self, attr):
         return getattr(self._the_tld, attr)
     def __setattr__(self, attr, val):
         setattr(self._the_tld, attr, val)
     def __delattr__(self, attr):
         delattr(self._the_tld, attr)

class MonkeyPopen(subprocess.Popen):
    def __init__(self, *args, **kwds):
        args = list(args)
        if len(args) or 'args' in kwds:
            for i, (k, d) in enumerate((('args', None), ('bufsize', -1), ('executable', None), ('stdin', None), ('stdout', None), ('stderr', None))):
                if len(args) <= i:
                    args.append(kwds.pop(k, d))
            if args[3] == None: args[3] = sys.stdin
            if args[4] == None: args[4] = sys.stdout
            if args[5] == None: args[5] = sys.stderr
            if args[5] == subprocess.STDOUT: args[5] = sys.stdout
        super().__init__(*args, **kwds)

def monkey_system(cmd):
    status = subprocess.call(cmd, shell=True)
    if status < 0: return -status
    else: return status * 256

def hook_stdio():
    sys.stdin = io.TextIOWrapper(TLDProxy(sys.stdin.buffer), line_buffering=True)
    sys.stdout = io.TextIOWrapper(TLDProxy(sys.stdout.buffer), line_buffering=True)
    sys.stderr = io.TextIOWrapper(TLDProxy(sys.stderr.buffer), line_buffering=True)
    subprocess.Popen = MonkeyPopen
    os.system = monkey_system

hook_stdio()

class RedirectSTDIO:
    def __init__(self, stdin, stdout, stderr):
        self.old_stdin = sys.stdin.buffer._the_tld
        self.old_stdout = sys.stdout.buffer._the_tld
        self.old_stderr = sys.stderr.buffer._the_tld
        self.new_stdin = stdin
        self.new_stdout = stdout
        self.new_stderr = stderr
    def __enter__(self):
        sys.stdin.buffer._tld.value = self.new_stdin
        sys.stdout.buffer._tld.value = self.new_stdout
        sys.stderr.buffer._tld.value = self.new_stderr
    def __exit__(self, *args):
        try:
            sys.stdin.flush()
            sys.stdout.flush()
            sys.stderr.flush()
        finally:
            sys.stdin.buffer._tld.value = self.old_stdin
            sys.stdout.buffer._tld.value = self.old_stdout
            sys.stderr.buffer._tld.value = self.old_stderr



import html
from ejcli.http import tasks, problem_info, contest_info
from ejcli.error import EJError

def split_tex(data):
    i = 0
    while True:
        j = data.find('$', i)
        if j < 0:
            yield ('normal', data[i:])
            return
        else:
            if j and data[j-1] == '\\':
                k = j
                while k and data[k-1] == '\\':
                    k -= 1
                if (j - k) % 2:
                    yield data[:k] + '\\'*((j - k)//2) + '$'
                    i = j + 1
                    continue
            yield ('normal', data[i:j])
            i = j
            while j < len(data) and data[j] == '$':
                j += 1
            if j == len(data):
                yield ('normal', data[i:j])
                return
            ss = j-i
            i = j
            brlevel = 0
            ss1 = 0
            while j < len(data) and (brlevel or ss1 < ss):
                if data[j] == '\\':
                    ss1 = 0
                    j += 2
                    continue
                elif data[j] == '$':
                    ss1 += 1
                    j += 1
                else:
                    ss1 = 0
                    if data[j] == '{':
                        brlevel += 1
                    elif data[j] == '}':
                        brlevel -= 1
                    j += 1
            if brlevel or ss1 < ss:
                yield ('normal', '$'*ss+data[i:])
                return
            elif max(map(ord, data[i:j])) >= 128:
                yield ('normal', '$'*ss+data[i:j])
                i = j
            else:
                yield ('tex', data[i:j-ss])
                i = j

def brmap(s):
    it = iter(enumerate(s))
    stack = []
    mapp = {}
    for i, c in it:
        if c == '\\':
            next(it)
        elif c == '{':
            stack.append(i)
        elif c == '}' and stack:
            j = stack.pop()
            mapp[i] = j
            mapp[j] = i
    return mapp

def match(s, bm, i, what):
    i0 = i
    ans = []
    j = 0
    while j < len(what) and i < len(s):
        if what[j] == ' ':
            while i < len(s) and s[i].isspace():
                i += 1
        elif what[j:j+2] == '\\ ':
            if s[i] != ' ':
                return None, i0
            i += 1
            j += 1
        elif what[j] == '{':
            if s[i] != '{' or i not in bm:
                ans.append(s[i])
                i += 1
            else:
                jj = bm[i]
                ans.append(s[i+1:jj])
                i = jj + 1
        else:
            if s[i] != what[j]:
                return None, i0
            i += 1
        j += 1
    if j < len(what):
        return None, i0
    return ans, i

def peek(s, bm, i):
    if s[i] == '{' and i in bm:
        return s[i+1:bm[i]], bm[i]+1
    return s[i], i+1

backslashes = {k: chr(v) for k, v in html.entities.name2codepoint.items()}
backslashes['dots'] = '...'
backslashes['ldots'] = '...'
backslashes['cdots'] = '...'
backslashes['cdot'] = backslashes['middot']
backslashes['rightarrow'] = '\u2192'
backslashes['bmod'] = ' mod '
backslashes['min'] = 'min'
backslashes['max'] = 'max'
backslashes['operatorname'] = ''
backslashes['mathbf'] = ''
backslashes['geq'] = backslashes['ge']
backslashes['leq'] = backslashes['le']
backslashes['neq'] = backslashes['ne']
backslashes['nless'] = '\u226e'
backslashes['ngtr'] = '\u226f'
backslashes['nleq'] = '\u2270'
backslashes['ngeq'] = '\u2271'
backslashes['ell'] = 'l'

def fn_underline(s, bm, i):
    q, i = peek(s, bm, i)
    ch = '\u0332'
    return ch + ch.join(q), i
backslashes['underline'] = fn_underline

def fn_limits(s, bm, i):
    q, i = match(s, bm, i, '_{^{')
    if q is None:
        return '', i
    return '[from ' + untex_expr(q[0]) + ' to ' + untex_expr(q[1]) + ']', i
backslashes['limits'] = fn_limits

def fn_text(s, bm, i):
    s, i = peek(s, bm, i)
    return untex(s), i
backslashes['text'] = fn_text
backslashes['mathrm'] = peek

def fn_xrightarrow(s, bm, i):
    args = []
    if s[i] == '[':
        j = s.find(']', i+1)
        if j >= 0:
            args.append(untex_expr(s[i+1:j]))
            i = j + 1
    ss, i = peek(s, bm, i)
    args.append(untex_expr(ss))
    return '\u2192 ['+', '.join(args)+']', i
backslashes['xrightarrow'] = fn_xrightarrow

def fn_frac(s, bm, i):
    q, i = match(s, bm, i, '{{')
    if q is None:
        return '', i
    return '('+untex_expr(q[0])+')/('+untex_expr(q[1])+')', i
backslashes['frac'] = fn_frac
backslashes['dfrac'] = fn_frac

def fn_not(s, bm, i):
    q, i = match(s, bm, i, '{')
    if q is None:
        return chr(824), i
    return chr(824)+untex_expr(q[0]).lstrip(), i
backslashes['not'] = fn_not

def fn_pmod(s, bm, i):
    q, i = match(s, bm, i, '{')
    if q is None:
        return '(mod)', i
    return '(mod '+untex_expr(q[0])+')', i
backslashes['pmod'] = fn_pmod
backslashes['enspace'] = '  '
backslashes['left'] = ''
backslashes['right'] = ''

def untex_expr(s):
    bm = brmap(s)
    ans = ''
    i = 0
    while i < len(s):
        if s[i] == '\\':
            if i + 1 >= len(s):
                break
            elif not s[i + 1].isalnum():
                ans += s[i + 1]
                i += 2
                continue
            else:
                j = i + 1
                while j < len(s) and s[j].isalnum():
                    j += 1
                cmd = s[i+1:j]
                i = j
                if cmd in backslashes:
                    if isinstance(backslashes[cmd], str):
                        ans += backslashes[cmd]
                    else:
                        q, i = backslashes[cmd](s, bm, i)
                        ans += q
                else:
                    ans += '\\' + cmd
        elif s[i] == '_' and i + 1 < len(s):
            ss, i = peek(s, bm, i + 1)
            ans += '[' + untex_expr(ss) + ']'
        elif s[i] == '{' or s[i] == '}':
            i += 1
            continue
        else:
            ans += s[i]
            i += 1
    return ans

def untex(s):
    ans = ''
    for kind, i in split_tex(s):
        if kind == 'normal':
            ans += i
        elif kind == 'tex':
            ans += ' '.join(untex_expr(i).split())
    return ans

def do_info(self, cmd):
    """
    usage: info [-t] [task]

    Show testing information and problem statement for a task, or for the whole contest if task is not specified.
    If -t is specified, attempt to decode embedded TeX expressions into human-readable form.
    """
    cmd = cmd.strip()
    tex = False
    if cmd[:2] == '-t' and (len(cmd) == 2 or cmd[2:3].isspace()):
        tex = True
        cmd = cmd[2:].strip()
    if cmd == '--help':
        return self.do_help('info')
    if cmd:
        try: task_id = next(i.id for i in tasks(self.url, self.cookie) if i.short_name == cmd)
        except StopIteration:
            raise EJError("No such task.")
        a, b = problem_info(self.url, self.cookie, task_id)
    else:
        b, a, j = contest_info(self.url, self.cookie)
        print(j)
    for k, v in a.items(): print(k+': '+v)
    print()
    print(untex(b) if tex else b)



def check_exists(file, options=set()):
    return True

def read_file(name, options=set()):
    return name

def get_fmt_name(s):
    s = s.split()
    for i in s:
        if i.startswith('__%'):
            return '%'+i[3:]
    spec = 'd'
    if 'unsigned' in s: spec = 'u'
    if 'char' in s: spec = 'c'
    if 'char' in s and '*' in s or 'char*' in s: spec = 's'
    if 'float' in s: spec = 'f'
    if 'double' in s: spec = 'lf'
    if '__char' in s: spec = 'c'
    if '__int' in s: spec = 'd'
    if '__unsigned' in s: spec = 'u'
    if '__hex' in s: spec = 'x'
    mod = ''
    if 'char' in s and spec not in ('c', 's'): mod = 'hh'
    else: mod += 'h'*s.count('short')+'l'*s.count('long')
    return '%'+mod+spec

def get_spec_name(s):
    return ' '.join(i for i in s.split() if i not in ('__char', '__int', '__unsigned', '__hex') and not i.startswith('__%'))

def format(s, options=set(), stdio=True):
    "type1 a, type2 b -> type_ans1, type_ans2: f1(a, b), f2(a, b)"
    args, exprs = s.split(':', 1)
    args, retvals = args.rsplit('->', 1)
    args = [i.strip() for i in args.split(',')]
    retvals = [i.strip() for i in retvals.split(',')]
    exprs = exprs.strip()
    code = '''\
#include <stdio.h>

''' if stdio else ''
    code += '''\
int main(void)
{
'''
    for i in args:
        code += '    '+get_spec_name(i.rsplit(' ', 1)[0])+' '+i.rsplit(' ', 1)[1]+';\n'
    code += '    scanf("'
    for i in args:
        code += get_fmt_name(i.rsplit(' ', 1)[0])
    code += '"'
    for i in args:
        code += ', &'+i.rsplit(' ', 1)[1]
    code += ');\n'
    if len(retvals) == 1 and get_fmt_name(retvals[0]) == '%s':
        code += '    puts('+exprs+');\n'
    else:
        code += '    printf("'
        b = False
        for i in retvals:
            if b: code += ' '
            else: b = True
            code += get_fmt_name(i)
        code += '\\n", '+exprs+');\n'
    code += '''\
    return 0;
}
'''
    return code



import socket, threading, os.path, pty, tty, select, sys, shlex, subprocess, io, signal, functools

mustexit = None

class UnbufferedStream(io.FileIO):
    def read1(self, x=-1): return self.read(x)

def readline(sock):
    ans = b''
    while not ans.endswith(b'\n'):
        try: chunk = sock.recv(1)
        except socket.error: return ''
        ans += chunk
        if chunk == b'': return ''
    return ans.decode('utf-8')[:-1]

def io_server_thread(sock, stdin, stdout, stderr, efd):
    try:
        sock.sendall(('pid %d\n'%os.getpid()).encode('utf-8'))
        fd_set = {sock.fileno(), stdout, stderr, efd}
        while True:
            l = select.select(list(fd_set), [], [])[0]
            if l == [efd]:
                sock.sendall(('exit %d\n'%os.read(efd, 1)[0]).encode('utf-8'))
                sock.close()
                break
            for i in l:
                if i == sock.fileno():
                    data = sock.recv(1048576, socket.MSG_PEEK)
                    if not data: fd_set.remove(i)
                    l = os.write(stdin, data)
                    sock.recv(l)
                    if os.isatty(stdin) and b'\3' in data[:l]: os.kill(os.getpid(), signal.SIGINT)
                elif i == stdout:
                    try: data = os.read(stdout, 1024)
                    except OSError: data = b''
                    if not data: fd_set.remove(i)
                    sock.sendall(('stdout %d\n'%len(data)).encode('utf-8')+data)
                elif i == stderr:
                    try: data = os.read(stderr, 1024)
                    except OSError: data = b''
                    if not data: fd_set.remove(i)
                    sock.sendall(('stderr %d\n'%len(data)).encode('utf-8')+data)
    except socket.error: pass
    finally:
        for i in {stdin, stdout, stderr, efd}: os.close(i)

tld_devtty = threading.local()

def torsocks_workaround():
    lp = os.getenv('LD_PRELOAD')
    if lp == None or not any('torsocks' in os.path.basename(i) for i in lp.split(':')):
        return socket.socket.connect, 'no_torsocks'
    un = os.uname()
    arch = un.machine
    is_x86 = len(arch) == 4 and arch.startswith('i') and arch.endswith('86') and arch[1] in '3456'
    is_x86_64 = arch == 'x86_64'
    if un.sysname != 'Linux' or not (is_x86 or is_x86_64) or os.path.isdir('/system'):
        print('Running ejcli --bash under torsocks is not supported on your system.')
        exit(1)
    import ctypes, mmap
    libc = ctypes.CDLL('libc.so.6')
    libc.mmap.restype = ctypes.POINTER(ctypes.c_char)
    libc.mmap.argtypes = (ctypes.c_void_p, ctypes.c_size_t, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_longlong)
    errno = ctypes.c_int.in_dll(libc, 'errno')
    ram = libc.mmap(None, 4096, mmap.PROT_READ|mmap.PROT_WRITE|4, mmap.MAP_PRIVATE|mmap.MAP_ANONYMOUS, -1, 0)
    if ctypes.cast(ram, ctypes.c_void_p).value == ctypes.c_void_p(-1).value:
        ctypes.pythonapi.PyErr_SetFromErrno(ctypes.py_object(OSError))
    if is_x86:
        # push ebx
        # mov eax, __NR_socketcall
        # mov ebx, [esp+8]
        # mov ecx, [esp+12]
        # int 0x80
        # pop ebx
        # ret
        for i, x in enumerate(b'\x53\xb8\x66\x00\x00\x00\x8b\x5c\x24\x08\x8b\x4c\x24\x0c\xcd\x80\x5b\xc3'): ram[i] = x
        _socketcall = ctypes.cast(ram, ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, ctypes.c_char_p))
        def socketcall(call, *args):
            param = b''
            for i in args:
                sz = ctypes.sizeof(i)
                ptr = ctypes.cast(ctypes.pointer(i), ctypes.POINTER(ctypes.c_char*sz))
                data = ptr[0][:]
                data += bytes(ctypes.sizeof(ctypes.c_long) - len(data))
                param += data
            return _socketcall(call, param)
        def _connect(fd, addr, sz):
            return socketcall(3, ctypes.c_int(fd), ctypes.c_char_p(addr), ctypes.c_size_t(sz))
    elif is_x86_64:
        # mov eax, __NR_connect
        # syscall
        # ret
        for i, x in enumerate(b'\xb8\x2a\x00\x00\x00\x0f\x05\xc3'): ram[i] = x
        _connect = ctypes.cast(ram, ctypes.CFUNCTYPE(ctypes.c_int, ctypes.c_int, ctypes.c_char_p, ctypes.c_size_t))
    else: assert False
    def connect(sock, addr):
        fd = sock.fileno()
        host, port = addr
        host = socket.inet_aton(host)
        port = port.to_bytes(2, 'big')
        addr = socket.AF_INET.to_bytes(2, 'little') + port + host
        addr += bytes(16 - len(addr))
        ans = _connect(fd, addr, len(addr))
        if ans < 0:
            errno.value = -ans
            ctypes.pythonapi.PyErr_SetFromErrno(ctypes.py_object(socket.error))
    return connect, 'torsocks'

sock_connect, torsocks_type = torsocks_workaround()

def io_server(brute, sock, auth_token, tty_conf):
    auth = readline(sock)
    if auth.count(':') != 1:
        sock.close()
        return
    ts_type, auth = auth.split(':')
    if auth != auth_token:
        sock.close()
        return
    if ts_type != torsocks_type:
        msg = b'Run the whole ejcli session under torsocks!\r\n'
        sock.sendall(('stdout %d\n'%len(msg)).encode('ascii')+msg+b'exit \0')
        sock.close()
        return
    mode, command = readline(sock).split(' ', 1)
    if mode == 'pty':
        stdin, cstdin = pty.openpty()
        tty.tcsetattr(cstdin, tty.TCSAFLUSH, tty_conf)
        stdout = stderr = stdin
        cstdout = cstderr = cstdin
        tld_devtty.value = '/proc/self/fd/%d'%cstdin
    else:
        cstdin, stdin = os.pipe()
        stdout, cstdout = os.pipe()
        stderr, cstderr = os.pipe()
    command = eval(command)
    efd, efd_w = os.pipe()
    threading.Thread(target=io_server_thread, args=(sock, stdin, stdout, stderr, efd), daemon=True).start()
    if mode == 'pty':
        do_open = UnbufferedStream
    else:
        do_open = open
    cstdin_f = do_open(cstdin, 'rb', closefd=False)
    cstdout_f = do_open(cstdout, 'wb', closefd=False)
    cstderr_f = do_open(cstderr, 'wb', closefd=False)
    from ejcli.hook_stdio import RedirectSTDIO
#   sys.stdin.buffer._tld.value = cstdin_f
#   sys.stdout.buffer._tld.value = cstdout_f
#   sys.stderr.buffer._tld.value = cstderr_f
    exitstatus = 0
    try:
        with RedirectSTDIO(cstdin_f, cstdout_f, cstderr_f):
            brute.onecmd(command)
    except SystemExit as e:
        if isinstance(brute, ForkingBrute): #we're in fork
            raise
        if len(e.args) == 1 and isinstance(e.args[0], int):
            exitstatus = e.args[0]
        elif len(e.args) == 1:
            print(e.args[0], file=sys.stderr)
        else:
            print(e.args, file=sys.stderr)
    except:
        if mustexit != None: exit(mustexit)
        sys.excepthook(*sys.exc_info())
        exitstatus = 1
    finally:
#       sys.stdin.flush()
#       sys.stdout.flush()
#       sys.stderr.flush()
#       del sys.stdin.buffer._tld.value
#       del sys.stdout.buffer._tld.value
#       del sys.stderr.buffer._tld.value
        try: del tld_devtty.value
        except AttributeError: pass
        for i in {cstdin, cstdout, cstderr}: os.close(i)
    try:
        os.write(efd_w, bytes((exitstatus,)))
        os.close(efd_w)
    except OSError: pass

def io_server_main(brute, sock, auth_token):
    tty_conf = tty.tcgetattr(0)
    while True: io_server(brute, sock.accept()[0], auth_token, tty_conf)

def hook_stdio():
    import ejcli.hook_stdio
    import os
    os_open = os.open
    def gp_open(file, *args, **kwds):
        if file == '/dev/tty' and hasattr(tld_devtty, 'value'):
            file = tld_devtty.value
        return os_open(file, *args, **kwds)
    os.open = gp_open

def start_io_server(brute, auth_token, zsh=False):
    import socket
    sock = socket.socket()
    sock.bind(('127.0.0.1', 0))
    sock.listen(1)
    threading.Thread(target=run_bash, args=(sock.getsockname(), auth_token, zsh), daemon=True).start()
    try: io_server_main(brute, sock, auth_token)
    except:
        if mustexit != None: exit(mustexit)
        raise

def run_bash(arg, auth_token, zsh=False):
    global mustexit
    port = arg[1]
    env = dict(os.environ)
    ppath = os.environ.get('PYTHONPATH', None)
    ppath = os.path.split(os.path.split(__file__)[0])[0]+(':'+ppath if ppath != None else '')
    env['PYTHONPATH'] = ppath
    func_code = '() { python3 -m ejcli.bashhelper %d %s "$@"; }'%(port, auth_token)
    print('bash integration enabled. Type `bj help` for help.')
    import tempfile
    if zsh:
        dotdir = env.get('ZDOTDIR', env['HOME'])
        with tempfile.TemporaryDirectory() as dir:
            with open(dir+'/.zshrc', 'w') as file:
                file.write('source '+shlex.quote(dotdir+'/.zshrc')+'\n')
                file.write('command_not_found_handler '+func_code+'\n\n')
                file.write('bj '+func_code+'\n')
            env['ZDOTDIR'] = dir
            mustexit = subprocess.call('zsh', env=env)
    else:
        with tempfile.NamedTemporaryFile(mode='w') as file:
            file.write('source '+shlex.quote(env['HOME']+'/.bashrc')+'\n')
            file.write('command_not_found_handle '+func_code+'\n\n')
            file.write('bj '+func_code+'\n')
            file.flush()
            mustexit = subprocess.call(('bash', '--rcfile', file.name), env=env)
    os.kill(os.getpid(), signal.SIGINT)

def io_client(port, auth_token, cmd):
    sock = socket.socket()
    sock_connect(sock, ('127.0.0.1', port))
    sock.sendall((torsocks_type+':'+auth_token+'\n').encode('utf-8'))
    if sys.stdin.isatty() and sys.stdout.isatty() and sys.stderr.isatty():
        mode = 'pty'
    else:
        mode = 'pipe'
    sock.sendall(('%s %r\n'%(mode, cmd)).encode('utf-8'))
    if mode == 'pty':
        old = tty.tcgetattr(0)
        tty.setraw(0)
    try:
        pid = None
        have_int = [False, False]
        def handler(*args):
            if have_int[1]:
                have_int[1] = False
                raise KeyboardInterrupt
            have_int[0] = True
        signal.signal(signal.SIGINT, handler)
        while True:
            try:
                have_int[1] = True
                if have_int[0]:
                    have_int[0] = False
                    have_int[1] = False
                    raise KeyboardInterrupt
                r = select.select([sys.stdin.fileno(), sock.fileno()], [], [])[0]
                have_int[1] = False
            except KeyboardInterrupt:
                if pid == None: raise
                os.kill(pid, signal.SIGINT)
            for i in r:
                if i == sys.stdin.fileno():
                    sock.sendall(sys.stdin.buffer.raw.read(1048576))
                elif i == sock.fileno():
                    try: where, l = readline(sock).split()
                    except ValueError: break
                    l = int(l)
                    if where == 'exit':
                        sock.close()
                        exit(l)
                    elif where == 'pid':
                        pid = l
                        continue
                    data = b''
                    while len(data) < l: data += sock.recv(l - len(data))
                    try:
                        if where == 'stdout': sys.stdout.buffer.raw.write(data)
                        else: sys.stderr.buffer.raw.write(data)
                    except IOError: pass
            else: continue
            break
    finally:
        if mode == 'pty':
            tty.tcsetattr(0, tty.TCSAFLUSH, old)

def smart_quote(x):
    if not (set(' \\\'\"\n') & set(x)): return x
    return shlex.quote(x)

class ForkingBrute:
    def __init__(self):
        import pickle
        self.brute = pickle.dumps(None)
        self.unpickle_failed = False
        self.execute(self.init)
        if self.brute == pickle.dumps(None):
            raise RuntimeError("ForkingBrute failed to initialize")
    def init(self):
        import ejcli.cmd
        self.brute = ejcli.cmd.BruteCMD()
        self.brute.stdin = self.brute.stdout = None
    def do_onecmd(self, cmd):
        self.brute.stdin = sys.stdin
        self.brute.stdout = sys.stdout
        self.brute.onecmd(cmd)
        self.brute.stdin = self.brute.stdout = None
    def execute(self, fn, *args):
        import pickle
        pipe = os.pipe()
        pid = os.fork()
        if not pid:
            try:
                os.close(pipe[0])
                try: self.brute = pickle.loads(self.brute)
                except:
                    self.unpickle_failed = True
                    raise
                fn(*args)
                with open(pipe[1], 'wb') as file: file.write(pickle.dumps(self.brute))
                sys.exit(0)
            except SystemExit: raise
            except BaseException:
                sys.excepthook(*sys.exc_info())
                with open(pipe[1], 'wb') as file: file.write(self.brute if self.unpickle_failed else pickle.dumps(self.brute))
                sys.exit(1)
        os.close(pipe[1])
        os.waitpid(pid, 0)
        with open(pipe[0], 'rb') as file: self.brute = file.read()
    def onecmd(self, cmd):
        self.execute(self.do_onecmd, cmd)

def main():
    signal.signal(signal.SIGTTIN, signal.SIG_IGN)
    signal.signal(signal.SIGTTOU, signal.SIG_IGN)
    if len(sys.argv) == 1 or sys.argv[1] in ('--bash', '--zsh'):
        hook_stdio()
        if 'BJ_PICKLE' in os.environ:
            brute = ForkingBrute()
        else:
            import ejcli.cmd
            brute = ejcli.cmd.BruteCMD()
        auth_token = '%0128x'%int.from_bytes(os.urandom(64), 'big')
        start_io_server(brute, auth_token, zsh=(len(sys.argv) > 1 and sys.argv[1] == '--zsh'))
    else:
        io_client(int(sys.argv[1]), sys.argv[2], ' '.join(map(smart_quote, sys.argv[3:])))

if __name__ == '__main__':
    main()



from .oauth_old import do_oauth as do_oauth_old
from .oauth_phjs import do_oauth as do_oauth_phjs

def do_oauth(url, email, password):
    if url.endswith('#oauth-old'):
        return do_oauth_old(email, password)
    return do_oauth_phjs(email, password)



import urllib.request, urllib.parse, json, collections, html
from .ejudge import Ejudge
from .base import Backend
from .openerwr import OpenerWrapper
from ..error import EJError
import ejcli._http.html2md as html2md
import ejcli._http.types as bjtypes

SUBM_LIST_URL = "/py/problem/0/filter-runs?problem_id=0&from_timestamp=-1&to_timestamp=-1&group_id=0&user_id=%s&lang_id=-1&status_id=-1&statement_id=&count=%d&with_comment=&page=%d"

class Informatics(Ejudge):
    @staticmethod
    def detect(url):
        for proto in ('http', 'https'):
            for domain in ('mccme', 'msk'):
                if (url+'/').startswith('%s://old.informatics.%s.ru/'%(proto, domain)):
                    for path in ('cgi-bin', 'ej'):
                        if (url+'/').startswith('%s://old.informatics.%s.ru/%s/'%(proto, domain, path)):
                            break
                    else:
                        return True
        return False
    def __init__(self, url, login, password):
        Backend.__init__(self)
        for proto in ('http', 'https'):
            for domain in ('mccme', 'msk'):
                if (url+'/').startswith('%s://old.informatics.%s.ru/'%(proto, domain)):
                    for vt in (('', ''), ('', '3'), ('moodle/', ''), ('moodle/', '3')):
                        if url.startswith('%s://old.informatics.%s.ru/%smod/statements/view%s.php?'%((proto, domain)+vt)):
                            if not vt[1]:
                                url = url.replace('view3', 'view', 1)
                            if vt[0]:
                                url = url.replace('/moodle/', '/', 1)
                            break
                    else:
                        raise EJError("URL must point to a contest or a task.")
                    if domain == 'mccme':
                        url = url.replace('mccme', 'msk', 1)
                    break
            else: continue
            break
        else:
            raise EJError("Not an old.informatics.msk.ru URL")
        self.url = url
        self.opener = OpenerWrapper(urllib.request.build_opener(urllib.request.HTTPCookieProcessor))
        self.opener.addheaders = [('User-Agent', 'ejcli')]
        self.opener.open("https://old.informatics.msk.ru/")
        req = self.opener.open("https://old.informatics.msk.ru/login/index.php", urllib.parse.urlencode({
            'username': login,
            'password': password,
            'testcookies': 1
        }).encode('ascii'))
        if req.geturl() != 'https://old.informatics.msk.ru/':
            raise EJError("Login failed.")
        self.registered = False
        data = self.opener.open(url).read().decode('utf-8', 'replace')
        query = dict(tuple(i.split('=', 1)) for i in url.split('#', 1)[0].split('?', 1)[1].split('&'))
        if 'id' in query and '<form action="view.php?id='+query['id']+'&register=1" method="post">' in data:
            return
        self._init_full(data)
    def _init_full(self, data=None):
        if data == None:
            data = self.opener.open(url).read().decode('utf-8', 'replace')
        self.user_id = int(data.split('<div id="user_id" style="display: none">', 1)[1].split('</div>', 1)[0])
        query = dict(tuple(i.split('=', 1)) for i in self.url.split('#', 1)[0].split('?', 1)[1].split('&'))
        if 'id' in query:
            tasks = []
            cur_task = int(data.split('<div id="problem_id" style="display: none">', 1)[1].split('</div>', 1)[0])
            data = data.split('<div class="statements_toc_alpha"><font size="-1"><ul>', 1)[1].split('</ul>', 1)[0]
            for i in data.split('<li>')[1:]:
                i = i.split('</li>', 1)[0]
                if i.startswith('<a href="view.php?'):
                    tasks.append(int(i.split('chapterid=', 1)[1].split('"', 1)[0].split('&', 1)[0]))
                else:
                    tasks.append(cur_task)
        else:
            tasks = [int(query['chapterid'])]
        self.task_list = tasks
        self.subm_list = []
        self.subm_list_partial = []
        self.subm_set = {}
        self._cache = {}
        self.registered = True
        self.submissions()
    def _cache_get(self, url):
        with self.cache_lock:
            if url in self._cache: return self._cache[url]
        ans = self.opener.open("https://old.informatics.msk.ru"+url).read()
        with self.cache_lock:
            if self.caching: self._cache[url] = ans
        return ans
    def _request_json(self, url):
        return json.loads(self._cache_get(url).decode('utf-8', 'replace'))
    def tasks(self):
        if not self.registered: return []
        return [bjtypes.task_t(i, str(i), None) for i in self.task_list]
    def submissions(self):
        if not self.registered: return []
        subms = []
        subms_partial = []
        page_cnt = self._request_json(SUBM_LIST_URL%(self.user_id, 100, 1))['metadata']['page_count']
        for i in range(page_cnt):
            data = self._request_json(SUBM_LIST_URL%(self.user_id, 100, i+1))
            if 'data' not in data: break #FIXME
            data = data['data']
            for i in data:
                task_id = i['problem']['id']
                si = i['id']
                if si in self.subm_set: break
                status = self._submission_status(i)
                score = self._submission_score(i)
                oktests = i.get('ejudge_test_num', None)
                subms.append([si, task_id, status, score, oktests])
                if int(task_id) in self.task_list: subms_partial.append([si, task_id, status, score, oktests])
            else: continue
            break
        idx = -len(self.subm_list)-1
        max_subm = -1
        for a, b, c, d, e in reversed(subms):
            if c[1] < 0: break
            self.subm_set[a] = idx
            idx -= 1
            max_subm = max(max_subm, a)
        slp = subms_partial + self.subm_list_partial
        self.subm_list[:0] = (i for i in subms if i[0] <= max_subm)
        self.subm_list_partial[:0] = (i for i in subms_partial if i[0] <= max_subm)
        return [bjtypes.submission_t(i, j, k[0], l, m) for i, j, k, l, m in slp]
    def submission_protocol(self, id):
        data = self._request_json("/py/protocol/get/%d"%int(id)).get('tests', {})
        ans = []
        status_codes = dict(
            OK='OK',
            RT='Run-time error',
            TL='Time-limit exceeded',
            PE='Presentation error',
            WA='Wrong answer',
            CF='Check failed',
            ML='Memory limit exceeded',
            SE='Security violation',
            SV='Style violation',
            WT='Wall time limit exceeded',
            SK='Skipped'
        )
        return [bjtypes.test_t(status_codes[data[str(k)]['status']], {'time_usage': data[str(k)]['time']/1000, 'memory_usage': int(data[str(k)]['max_memory_used'])}) for k in sorted(map(int, data))]
    def submit_solution(self, task, lang, text):
        if not self.registered: return
        if isinstance(text, str): text = text.encode('utf-8')
        data = []
        data.append(b'"lang_id"\r\n\r\n'+str(lang).encode('ascii'))
        data.append(b'"file"; filename="brute.txt"\r\nContent-Type'
                    b': text/plain\r\n\r\n'+text)
#       data.append(b'"action_40"\r\n\r\nSend!')
        import random
        while True:
            x = b'----------'+str(random.randrange(1, 1000000000)).encode('ascii')
            for i in data:
                if x in i: break
            else: break
    #   x = '-----------------------------850577185583170701784494929'
        data = b'\r\n'.join(b'--'+x+b'\r\nContent-Disposition: form-data; name='+i for i in data)+b'\r\n--'+x+b'--\r\n'
        self.opener.open(urllib.request.Request("https://old.informatics.msk.ru/py/problem/%d/submit"%task, data, {'Content-Type': 'multipart/form-data; boundary='+x.decode('ascii')}))
        with self.cache_lock: self.stop_caching()
    def status(self):
        if not self.registered: return {}
        self.submissions()
        by_task = {}
        for a, b, c, d, e in self.subm_list_partial:
            if b not in by_task or by_task[b][1] < c[1]:
                by_task[int(b)] = c
        return collections.OrderedDict((str(i), by_task.get(i, (None,))[0]) for i in self.task_list)
    def scores(self):
        if not self.registered: return {}
        self.submissions()
        by_task = {}
        for a, b, c, d, e in self.subm_list_partial:
            if d != None and (b not in by_task or by_task[b] < d):
                by_task[int(b)] = d
        return collections.OrderedDict((str(i), by_task.get(i, None)) for i in self.task_list)
    def compile_error(self, id):
        data = self._request_json("/py/protocol/get/%d"%int(id))
        return data.get('compiler_output', data.get('protocol', '')) or None
    def _submission_status(self, obj):
        statuses = {
            0: ('OK', 2),
            98: ('Compiling...', -1),
            96: ('Running...', -1),
            1: ('Compilation error', 0),
            7: ('Partial solution', 1),
            377: ('In queue', -1),
        }
        status = obj['ejudge_status']
        return statuses.get(status, ('Unknown status #%d'%status, 0))
    def submission_source(self, id):
        data = self._request_json("/py/problem/run/%d/source"%int(id))
        return data.get('data', {}).get('source', None).encode('utf-8')
    def do_action(self, action, *args):
        if action == 'start_virtual':
            if self.registered: return False
            data = self.opener.open(self.url).read().decode('utf-8', 'replace')
            query = dict(tuple(i.split('=', 1)) for i in self.url.split('#', 1)[0].split('?', 1)[1].split('&'))
            if 'id' not in query or '<form action="view.php?id='+query['id']+'&register=1" method="post">' not in data:
                return False
            data = self.opener.open('https://old.informatics.msk.ru/mod/statements/view.php?id='+query['id']+'&register=1', b'').read().decode('utf-8', 'replace')
            if '<script type="text/javascript">\n//<![CDATA[\n\n  function redirect() {\n      document.location.replace(\'view.php?id='+query['id']+'\');\n  }\n  setTimeout("redirect()", 3000);\n//]]>\n</script>' not in data: return False
            self._init_full()
        raise EJError("NYI")
    def compiler_list(self, prob_id):
        known_compilers = {1: 'fpc', 2: 'gcc', 3: 'g++', 22: 'php', 23: 'python', 24: 'perl', 25: 'mcs', 26: 'ruby', 27: 'python3'}
        data = self._cache_get("/mod/statements/view.php?chapterid=%d"%prob_id).decode('utf-8', 'replace')
        ans = []
        for i in data.split('<select name="lang_id" id="lang_id" ', 1)[1].split('>', 1)[1].split('</select>', 1)[0].replace('<option value="', "<option value='").split("<option value='")[1:]:
            a = int(i.split("'", 1)[0].split('"', 1)[0])
            c = html.unescape(i.split('</option>', 1)[0].split('>', 1)[1])
            b = known_compilers.get(a, str(a))
            ans.append(bjtypes.compiler_t(a, b, c.strip()))
        return ans
    def submission_stats(self, id):
        data = self._request_json("/py/protocol/get/%d"%int(id))
        ans = {}
        score = [i[3] for i in self.subm_list_partial if i[0] == id]
        if len(score) == 1 and score[0] != None:
            ans['score'] = score[0]
        if 'tests' in data:
            ans['tests'] = {}
            ans['tests']['total'] = len(data['tests'])
            ans['tests']['success'] = sum(1 for k, v in data['tests'].items() if v['status'] == 'OK')
            ans['tests']['fail'] = ans['tests']['total'] - ans['tests']['success']
        return (ans, None)
    def _submission_score(self, data):
        return data['ejudge_score'] if data['ejudge_score'] != None and data['ejudge_score'] >= 0 else None
    def contest_info(self):
        return ('', {}, {})
    def problem_info(self, id):
        url = "/mod/statements/view.php?chapterid=%d"%id
        data = self._cache_get(url).decode('utf-8', 'replace')
        if '<div class="legend">' in data: the_html = data.split('<div class="legend">', 1)[1]
        elif '<div class="statements_content">' in data:
            the_html = data.split('<div class="statements_content">', 1)[1]
            the_html = the_html.split('<h1>', 2)
            the_html = '<h1>'+the_html[0]+the_html[2]
        else: return ({}, None)
        return ({}, html2md.html2md(the_html.split("<div id='submit' ", 1)[0], None, "https://old.informatics.msk.ru"+url))
    def download_file(self, prob_id, filename):
        raise EJError("File download doesn't exist on old.informatics.msk.ru")
    def clar_list(self):
        raise EJError("Clarifications don't exits on old.informatics.msk.ru")
    def submit_clar(self, *args):
        raise EJError("Clarifications don't exits on old.informatics.msk.ru")
    def read_clar(self, id):
        raise EJError("Clarifications don't exits on old.informatics.msk.ru")
    def get_samples(self, subm_id):
        raise EJError("Can't get samples on old.informatics.msk.ru")
    def scoreboard(self):
        raise EJError("Scoreboard is not supported on old.informatics.msk.ru")
    def stop_caching(self):
        self._cache.clear()



from ejcli.http import locales, set_locale

def do_locale(self, cmd):
    """
    usage: locale [which]

    If called without arguments, list available locales.
    If called with an argument, set the current locale to <which>.
    """
    cmd = cmd.strip()
    if cmd:
        set_locale(self.url, self.cookie, cmd)
    else:
        print('Available locales:', ', '.join(i[1] for i in locales(self.url, self.cookie)))



import html, urllib.parse

def do_unescape(s, code=False):
    if code: return html.unescape(s)
    return html.unescape(' '.join(('_'+s+'_').split()))[1:-1]

def validate_tag(data):
    data = data.strip()
    if data.startswith('/'): data = data[1:]
    if data.endswith('/'): data = data[:-1]
    i = 0
    while i < len(data):
        j = min(data.find("'", i) % (len(data) + 1), data.find('"', i) % (len(data) + 1))
        if not set(data[i:j]).issubset(set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789- =')):
            return False
        if j == len(data): return True
        i = j + 1
        j = data.find(data[j], j + 1) % (len(data) + 1)
        if j == len(data):
            return False
        i = j + 1
    return i == len(data)

def html2md(data, dload_prefix=None, base=None):
    data = data.split('<')
    ans = do_unescape(data[0])
    is_code = False
    hrefs = []
    for i in data[1:]:
        if i.startswith('a href="'):
            href, i = i[8:].split('">', 1)
            href = do_unescape(href.split('"', 1)[0], is_code)
            if base != None: href = urllib.parse.urljoin(base, href)
            if dload_prefix != None and href.startswith(dload_prefix):
                href = 'file '+href[len(dload_prefix):]
            hrefs.append(href)
            ans += '['+do_unescape(i, is_code)
        elif i.startswith('img '):
            x, i = i.split('>', 1)
            src = x.split('src="', 1)[1].split('"', 1)[0]
            src = do_unescape(src.split('"', 1)[0], is_code)
            if base != None: src = urllib.parse.urljoin(base, src)
            if dload_prefix != None and src.startswith(dload_prefix):
                src = 'file '+src[len(dload_prefix):]
            ans += '![]('+src+')'
        elif any(i.startswith('h%d>'%x) for x in range(1, 7)):
            d, i = i.split('>', 1)
            ans += '\n\n'+'#'*int(d[1:])+' '+do_unescape(i, is_code)
        elif i.startswith('li>'):
            ans += '* ' + do_unescape(i[3:], is_code)
        elif any(i.startswith(x) for x in ('br>', 'br/>', '/h1>', '/h2>', '/h3>', '/h4>', '/h5>', '/h6>', '/p>', '/li>', '/div>')):
            ans += '\n\n' + do_unescape(i.split('>', 1)[1], is_code)
        elif i.startswith('/a>'):
            ans += ']('+hrefs.pop()+')'+do_unescape(i[3:], is_code)
        elif i.startswith('sub>') or i.startswith('sub '):
            ans += '[' + do_unescape(i.split('>', 1)[-1], is_code)
        elif i.startswith('/sub>'):
            ans += ']' + do_unescape(i[5:], is_code)
        elif i.startswith('sup>') or i.startswith('sup '):
            ans += '**(' + do_unescape(i.split('>', 1)[-1], is_code)
        elif i.startswith('/sup>'):
            ans += ')' + do_unescape(i[5:], is_code)
        else:
            if i.startswith('/pre>'):
                ans += '\n```'
                is_code = False
            elif i.startswith('/code>'):
                ans += '`'
                is_code = False
            elif i.startswith('pre>') or i.startswith('pre '):
                ans += '```\n'
                is_code = True
            elif i.startswith('code>') or i.startswith('code '):
                ans += '`'
                is_code = True
            if '>' not in i or not validate_tag(i.split('>', 1)[0]):
                ans += '<' + do_unescape(i, is_code)
            else:
                ans += do_unescape(i.split('>', 1)[-1], is_code)
    return '\n\n'.join(i for i in ans.split('\n\n') if i).replace('\n\n\n', '\n\n').strip()

def md2html(data):
    ans = []
    uldepth = 0
    for l in data.split('\n\n'):
        cur_uldepth = 0
        if set(l.split(' ', 1)[0]) == {'*'}:
            hdeep, l = l.split(' ', 1)
            cur_uldepth = len(hdeep)
        while uldepth < cur_uldepth:
            ans.append('<ul>')
            uldepth += 1
        while cur_uldepth < uldepth:
            ans.append('</ul>')
            uldepth -= 1
        l = l.replace('\n', ' ')
        fmt = '<p>%s</p>' if uldepth == 0 else '<li>%s</li>'
        if set(l.split(' ', 1)[0]) == {'#'}:
            hdeep, l = l.split(' ', 1)
            fmt = '<h%d>%%s</h%d>'%(len(hdeep), len(hdeep))
        l = l.split('[')
        ll = html.escape(l[0])
        for i in l[1:]:
            if ']' not in i:
                ll += html.escape(i)
                continue
            href, i = i.split(']', 1)
            text = href
            if i.startswith('(') and ')' in i:
                text, i = i[1:].split(')')
            ll += '<a href="'+html.escape(href)+'">'+html.escape(text)+'</a>'+html.escape(i)
        ans.append(fmt % ll)
    return '\n'.join(ans)



import sys, os.path

def read_file(name, options=set(), includes=None):
    name = os.path.realpath(name)
    if includes == None:
        includes = set()
    if name in includes: return ''
    includes.add(name)
    with open(name) as file:
        lines = file.read().split('\n')
    if '+include' in options:
        for i in range(len(lines)):
            l = lines[i]
            if l.startswith('#include "'):
                p2 = os.path.join(os.path.split(name)[0], eval(l[9:]))
                lines[i] = read_file(p2, includes=includes)+'\n'
    ans = '\n'.join(lines)
    ans2 = ''
    c = 0
    for i in ans:
        if i == '\n':
            c += 1
            if c >= 3: continue
        else:
            c = 0
        ans2 += i
    includes.remove(name)
    return ans2

def escape_strings(s):
    instr = None
    ans = ''
    for c in s:
        if instr == None:
            ans += c
            if c in ('"', "'"):
                instr = c
        elif instr.endswith('\\'):
            ans += '\\x%02x'%ord(c)
            instr = instr[:1]
        elif c == '\\':
            ans += '\\x%02x'%ord(c)
            instr += '\\'
        elif instr == c:
            instr = None
            ans += c
        else:
            ans += '\\x%02x'%ord(c)
    return ans

def unescape_strings(s):
    ans = ''
    i = 0
    while i < len(s):
        i1 = s.find('"', i)
        i2 = s.find("'", i)
        if i1 < 0: i1 = i2
        if i2 < 0: i2 = i1
        if i1 < 0:
            ans += s[i:]
            break
        j = min(i1, i2)
        ans += s[i:j]
        c = s[j]
        i = s.find(c, j+1)
        chars = ''.join(chr(int(i, 16)) for i in s[j+1:i].split('\\x')[1:])
        ans += c + chars + c
        i += 1
    return ans

def format(original, options=set(), cplusplus=True):
    original = escape_strings(original)
    i = 0
    s = [i.rstrip() for i in original.split('\n')]
    while i < len(s):
        if s[i] == '#pragma cut before':
            del s[:i+1]
            i = 0
            continue
        if s[i] in ('#pragma cut', '#pragma cut after'):
            del s[i:]
            break
        i += 1
    for i, j in enumerate(s):
        stripped = j.strip()
        if cplusplus and '+mods' in options and stripped in ('public:', 'private:', 'protected:'):
            l = j.find(stripped)
            while l % 4 != 1:
                l += 1
            s[i] = ' ' * l + stripped
        if (not cplusplus and '+funcvoid' in options) and j == stripped and j.endswith('()'):
            s[i] = j[:-2]+'(void)'
        if '+cexpr' in options and j.startswith('#pragma cexpr '):
            from .format_cexpr import format as format_cexpr
            s[i] = escape_strings(format_cexpr(unescape_strings(j[14:]), options, stdio=False).strip())
    s = [i.rstrip() for i in '\n'.join(s).split('\n')]
    i = 0
    while i < len(s):
        j = s[i]
        stripped = j.strip()
        if stripped != '{' and i and (s[i-1].endswith(')') or s[i-1].endswith('else') and not s[i-1][-5:-4].isalnum() and s[i-1][-5:-4] != '_') and ('+forcebraces' in options or s[i-1].strip() == '} else'):
            indent = s[i-1][:s[i-1].find(s[i-1].strip())]
            s.insert(i, indent+'{')
            ii = i + 1
            while ii < len(s) and s[ii].startswith(indent) and not s[ii][len(indent):len(indent)+1].strip():
                ii += 1
            s.insert(ii, indent+'}')
            continue
        if '+braces' in options:
            if stripped == '{' and i and not s[i-1].endswith('{'):
                s[i-1] += ' {'
                if s[i-1].strip() == 'else {':
                    s[i-1] = s[i-1].replace('else', '} else')
                    indent = s[i-1][:s[i-1].find(s[i-1].strip())]
                    ii = i - 2
                    while ii >= 0 and s[ii].startswith(indent) and not s[ii][len(indent):len(indent)+1].strip():
                        ii -= 1
                    if not ii:
                        s.insert(0, indent+'{')
                    else:
                        s[ii] += ' {'
                del s[i]
                continue
            elif stripped == 'else' and i and s[i-1].endswith('}'):
                s[i-1] += ' ' + stripped
                del s[i]
                continue
        i += 1
    s = '\n'.join(s)
    while ' \n' in s: s = s.replace(' \n', '\n')
    if '+ifspc' in options:
        i = 0
        while i < len(s):
            for j in ('if', 'while', 'for', 'switch'):
                if s[i:i+len(j)+1] == j + '(' and (not i or not s[i-1].isalnum() and s[i-1] != '_'):
                    s = s[:i+len(j)] + ' ' + s[i+len(j):]
                    break
            else: i += 1
    if '+pointers' in options:
        last_not_star = -1
        i = 0
        while i < len(s):
            if s[i] != '*':
                if last_not_star >= 0 and i - last_not_star > 1:
                    if not s[last_not_star].isspace() and s[i].isspace():
                        s = s[:last_not_star+1] + s[i] + s[last_not_star+1:i] + s[i+1:]
                    else:
                        last_not_star = i
                else:
                    last_not_star = i
            i += 1
    if '+funcgrep' in options:
        s = [i.rstrip() for i in s.split('\n')]
        for i, j in enumerate(s):
            if j[:1].isspace() or j.startswith('typedef') and not j[7:8].isalnum() and j[7:8] != '_':
                continue
            j = j.split('(')
            for k in range(1, len(j)):
                if j[k-1][-1:].isalnum() or j[k-1][-1:] == '_':
                    jj = len(j[k-1]) - 1
                    while jj >= 0 and (j[k-1][jj].isalnum() or j[k-1][jj] == '_'):
                        jj -= 1
                    jj += 1
                    if jj >= 0 and (jj > 0 or k == 1):
                        j[k-1] = j[k-1][:jj].rstrip() + '\n' + j[k-1][jj:]
            s[i] = '('.join(j)
        s = '\n'.join(s)
    return unescape_strings(s)



import collections, ast, html
from .libpcms import PCMS as LPCMS
from ejcli._http.base import Backend
from ejcli.error import EJError
import ejcli._http.html2md as html2md
import ejcli._http.types as bjtypes

class PCMS(Backend):
    @staticmethod
    def detect(url):
        return LPCMS.detect(url)
    def __init__(self, url, login, password):
        Backend.__init__(self)
        try: self.pcms = LPCMS(url, (login, password))
        except AssertionError as e: raise EJError(str(e))
        self._subms = []
        self._subms_set = set()
        self._clars = []
        self._messages = []
        self._user_clars = []
        self._user_clars_set = set()
        self.pcms.set_locale('English')
        self._call_cache = {}
        self.submission_list()
    @staticmethod
    def _remove_tags(t):
        t = t.split('<')
        return t[0] + ''.join(i.split('>', 1)[-1] for i in t[1:])
    def _cache_call(self, func, *args):
        with self.cache_lock:
            if (func, args) in self._call_cache: return self._call_cache[func, args]
        ans = func(*args)
        with self.cache_lock:
            if self.caching: self._call_cache[func, args] = ans
        return ans
    def tasks(self):
        try: return [bjtypes.task_t(i, j, None) for i, j in enumerate(self.pcms.get_submit()[0])]
        except: raise EJError("Failed to fetch task list.")
    def submissions(self):
        x = self._cache_call(self.pcms.get_submissions)
        x = list(sorted(x.items(), key=lambda x:x[0]))
        data = []
        sid_to_subm = {}
        for t, (k, v) in enumerate(x):
            for i in v:
                sid = int(i[1]['attempt']) * len(x) + t
                sid_to_subm[sid] = i
                data.append((int(i[1]['time'].replace(':', '')), int(i[1]['attempt']), sid, k))
        data.sort()
        for _, _, i, j in data:
            if i not in self._subms_set:
                self._subms_set.add(i)
                self._subms.append((i, j))
        ans = []
        for sid, task_id in reversed(self._subms):
            try: subm = sid_to_subm[sid]
            except KeyError: status = score = oktests = None
            else:
                status = self._remove_tags(subm['outcome']) if 'outcome' in subm else None
                if status == 'Accepted': status = 'OK'
                score = int(subm['score'].split(' = ', 1)[1]) if 'score' in subm else None
                ans.append(bjtypes.submission_t(sid, task_id, status, score, None))
        return ans
    def _decode_subm_id(self, subm_id):
        subm_id = int(subm_id)
        tasks = self.task_list()
        attempt, taskid = divmod(subm_id, len(tasks))
        task = tasks[taskid]
        return (task, attempt)
    def submission_protocol(self, subm_id):
        task, attempt = self._decode_subm_id(subm_id)
        proto = self._cache_call(self.pcms.get_protocol, task, attempt)
        if proto == None:
            raise EJError("Failed to fetch testing protocol.")
        ans = []
        for i in proto:
            if 'outcome' not in i[1]: continue
            outcome = self._remove_tags(i[1]['outcome'])
            if outcome == 'Accepted': outcome = 'OK'
            stats = {}
            try:
                stats['time_usage'] = int(i[1]['time'].split()[0]) / 1000
            except KeyError: pass
            if 'memory' in i[1]:
                m = i[1]['memory']
                if m.endswith('KB'): m = int(m[:2])*1024
                elif m.endswith('MB'): m = int(m[:2])*1048576
                else: assert False
                stats['memory_usage'] = m
            ans.append(bjtypes.test_t(outcome, stats))
        return ans
    def submit_solution(self, taskid, lang, text):
        tasks, langs, exts, vs = self._cache_call(self.pcms.get_submit)
        task = tasks[taskid]
        lang = [j for i, j, k in self.compiler_list(taskid) if i == lang][0]
        if isinstance(text, str):
            text = text.encode('utf-8')
        try: self.pcms.submit(task, lang, text, vs, {i[lang] for i in exts.values() if lang in i}.pop())
        except AssertionError: pass
        with self.cache_lock: self.stop_caching()
    def status(self):
        x = self._cache_call(self.pcms.get_submissions)
        ans = collections.OrderedDict()
        for k, v in x.items():
            if v:
                ans[k] = self._remove_tags(v[-1][1]['outcome']) if 'outcome' in v[-1][1] else None
            else:
                ans[k] = None
        return ans
    def scores(self):
        x = self._cache_call(self.pcms.get_submissions, False)
        ans = {}
        for k, v in x.items():
            notnone = False
            for i in v:
                if i[0].startswith('run '):
                    notnone = True
                    ans[k] = ans.get(k, 0)
                if i[0] == 'always' and 'head total-score' in i[1]:
                    ans[k] = int(i[1]['head total-score'].split('<span>Score =', 1)[1].split('</span>', 1)[0])
            if not notnone: ans[k] = None
        return ans
    def compile_error(self, subm_id):
        task, attempt = self._decode_subm_id(subm_id)
        ans = self._cache_call(self.pcms.get_compile_error, task, attempt)
        if ans == None: raise IndexError
        return ans
    def submission_source(self, subm_id):
        task, attempt = self._decode_subm_id(subm_id)
        return self._cache_call(self.pcms.get_source, task, attempt)
    def compiler_list(self, task_id):
        tasks, langs, jp, vs = self._cache_call(self.pcms.get_submit)
        return [bjtypes.compiler_t(i + 1, j, k) for i, (j, k) in enumerate(sorted(langs.items())) if j in jp[tasks[task_id]]]
    def submission_stats(self, subm_id):
        task, attempt = self._decode_subm_id(subm_id)
        ans = {}
        score = self.submission_score(subm_id)
        if score != None: ans['score'] = score
        proto = self._cache_call(self.pcms.get_protocol, task, attempt)
        if proto == None:
            raise EJError("Failed to fetch testing protocol.")
        group_scores = []
        passed = 0
        failed = 0
        for i in proto:
            if 'head group-score' in i[1]:
                score = int(i[1]['head group-score'].split(' = ', 1)[1].split('<', 1)[0])
                group_scores.append(score)
            if 'outcome' not in i[1]: continue
            if i[1]['outcome'] in ('Accepted', 'OK'): passed += 1
            else: failed += 1
        if passed + failed != 0:
            ans['tests'] = {}
            ans['tests']['total'] = passed + failed
            ans['tests']['success'] = passed
            ans['tests']['fail'] = failed
        if group_scores: ans['group_scores'] = group_scores
        return (ans, '')
    def download_file(self, *args):
        raise EJError("STUB")
    def do_action(self, *args):
        raise EJError("Not implemented on PCMS")
    def contest_info(self):
        props = self._cache_call(self.pcms.get_info)
        data1 = {}
        if 'Status' in props and '<script type="text/javascript">' in props['Status']:
            script = props['Status'].split('<script type="text/javascript">', 1)[1].split('</script>', 1)[0].strip()
            if script.startswith('clock.init('):
                qq = ast.literal_eval(script[10:-1]) # not a valid JSON
                if len(qq) == 5 and qq[2] == 'RUNNING':
                    data1['contest_time'] = qq[3] / 1000
                    data1['contest_duration'] = qq[4] / 1000
        datas = {k: html.unescape(v) for k, v in props.items() if '<' not in v}
        return ('', datas, data1)
    def problem_info(self, task_id):
        html, base = self._cache_call(self.pcms.get_links)
        return ({}, html2md.html2md(html, None, base))
    def _get_clars(self):
        tasks, clars, vs = self._cache_call(self.pcms.get_clars)
        self._clars = list(clars)
        return tasks, clars, vs
    def clar_list(self):
        self._get_clars()
        self._messages = [j for i, j in self._cache_call(self.pcms.get_messages) if i == 'msg']
        data = [(i * 2, j[1]['text'].split('.', 1)[0]+': '+j[1]['text pre'].replace('\r', '').replace('\n', ' ')+': '+j[1]['type']) for i, j in enumerate(reversed(self._clars))]
        data += [(i * 2 + 1, j) for i, j in enumerate(reversed(self._messages))]
        for x in data:
            if x[0] not in self._user_clars_set:
                self._user_clars_set.add(x[0])
                self._user_clars.append(x)
        return [bjtypes.clar_t(i[0], i[1]) for i in reversed(self._user_clars)]
    def submit_clar(self, task, subject, text):
        tasks, clars, vs = self._get_clars()
        #clars = list: type, text, 'text pre', type, 'answer pre'
        try: task = tasks[task]
        except IndexError: pass
        try: self.pcms.submit_clar(task, subject, text, vs)
        except AssertionError: raise EJError("Failed to submit clar")
        with self.cache_lock: self.stop_caching()
    def read_clar(self, id):
        self._get_clars()
        id, kind = divmod(id, 2)
        if kind == 0:
            return self._clars[-1-id][1]['answer pre']
        else:
            return self._messages[-1-id]
    def scoreboard(self):
        ans = []
        for i in self._cache_call(self.pcms.monitor)[1]:
            ans.append(({'name': i[0]['party']}, []))
            for j in i[1]:
                q = {}
                if j[1] and j[1][0] in ('+', '-'):
                    try: q['attempts'] = int(j[1][0]+'0'+j[1][1:])
                    except (TypeError, ValueError): pass
                else:
                    try: q['score'] = int(j[1])
                    except (TypeError, ValueError): pass
                if j[2]:
                    try:
                        a, b = j[2].split(':')
                        q['penalty'] = 60 * int(a) + int(b)
                    except (TypeError, ValueError): pass
                q['flags'] = list(j[0])
                ans[-1][1].append(q)
        return ans
    def stop_caching(self):
        self._call_cache.clear()



import subprocess
from ejcli.http import submission_source
from ejcli.error import EJError

def do_getsource(self, cmd):
    """
    usage: getsource <submission_id> <shell command>

    Retrieve the source code of a submission.
    """
    id, cmd = (cmd.strip()+' ').split(' ', 1)
    if not id.isnumeric():
        return self.do_help('getsource')
    src = submission_source(self.url, self.cookie, int(id))
    if src == None:
        raise EJError('Source code is not available')
    p = subprocess.Popen('cat '+cmd, stdin=subprocess.PIPE, shell=True)
    p.stdin.write(src)
    p.stdin.close()
    p.wait()
